#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitaions under the License.
#//----------------------------------------------------------------------

import cocotb
from cocotb.triggers import RisingEdge, FallingEdge, Timer

from uvm.base import UVMObject, sv, UVMConfigDb, UVM_HIGH
from uvm.macros import uvm_component_utils, uvm_object_utils
from uvm.comps import UVMMonitor
from uvm.tlm1 import UVMAnalysisPort

from ubus_transfer import *

#//------------------------------------------------------------------------------
#//
#// CLASS: slave_address_map_info
#//
#// The following class is used to determine which slave should respond
#// to a transfer on the bus
#//------------------------------------------------------------------------------

#class slave_address_map_info extends uvm_object
class slave_address_map_info(UVMObject):

    def __init__(self, name="slave_address_map_info"):
        UVMObject.__init__(self, name)
        self.min_addr = 0
        self.max_addr = 0


    def set_address_map(self, min_addr, max_addr):
        self.min_addr = min_addr
        self.max_addr = max_addr

    def get_min_add(self):
        return self.min_addr

    def get_max_add(self):
        return self.max_addr


uvm_object_utils(slave_address_map_info)
#  `uvm_object_utils_begin(slave_address_map_info)
#    `uvm_field_int(min_addr, UVM_DEFAULT)
#    `uvm_field_int(max_addr, UVM_DEFAULT)
#  `uvm_object_utils_end


#// Enumerated for ubus bus state

#typedef enum {RST_START, RST_STOP, NO_OP, ARBI, ADDR_PH, ADDR_PH_ERROR,
#  DATA_PH} ubus_bus_state
RST_START = 0
RST_STOP = 1
NO_OP = 2
ARBI = 3
ADDR_PH = 4
ADDR_PH_ERROR = 5
DATA_PH = 6


#//------------------------------------------------------------------------------
#//
#// CLASS: ubus_status
#//
#//------------------------------------------------------------------------------

class ubus_status(UVMObject):
    #
    def __init__(self, name="ubus_status"):
        UVMObject.__init__(self, name)
        self.bus_state = RST_START

uvm_object_utils(ubus_status)

#  `uvm_object_utils_begin(ubus_status)
#    `uvm_field_enum(ubus_bus_state, bus_state, UVM_DEFAULT)
#  `uvm_object_utils_end
#


#//------------------------------------------------------------------------------
#//
#// CLASS: ubus_bus_monitor
#//
#//------------------------------------------------------------------------------
#class ubus_bus_monitor extends uvm_monitor

class ubus_bus_monitor(UVMMonitor):
    #
    #  // The virtual interface used to view HDL signals.
    #  protected virtual ubus_if vif
    #
    #  // Property indicating the number of transactions occuring on the ubus.
    #  protected int unsigned num_transactions = 0
    #
    #  // The following two bits are used to control whether checks and coverage are
    #  // done both in the bus monitor class and the interface.
    #  bit checks_enable = 1
    #  bit coverage_enable = 1
    #
    #  // Analysis ports for the item_collected and state notifier.
    #  uvm_analysis_port #(ubus_transfer) item_collected_port
    #  uvm_analysis_port #(ubus_status) state_port
    #
    #  // The state of the ubus
    #  protected ubus_status status
    #
    #  // The following property holds the transaction information currently
    #  // being captured (by the collect_address_phase and data_phase methods).
    #  protected ubus_transfer trans_collected
    #
    #  // Events needed to trigger covergroups
    #  protected event cov_transaction
    #  protected event cov_transaction_beat
    #
    #  // Fields to hold trans data and wait_state.  No coverage of dynamic arrays.
    #  protected bit [15:0] addr
    #  protected bit [7:0] data
    #  protected int unsigned wait_state
    #
    #  // Transfer collected covergroup
    #  covergroup cov_trans @cov_transaction
    #    option.per_instance = 1
    #    trans_start_addr : coverpoint trans_collected.addr {
    #      option.auto_bin_max = 16; }
    #    trans_dir : coverpoint trans_collected.read_write
    #    trans_size : coverpoint trans_collected.size {
    #      bins sizes[] = {1, 2, 4, 8}
    #      illegal_bins invalid_sizes = default; }
    #    trans_addrXdir : cross trans_start_addr, trans_dir
    #    trans_dirXsize : cross trans_dir, trans_size
    #  endgroup : cov_trans
    #
    #  // Transfer collected data covergroup
    #  covergroup cov_trans_beat @cov_transaction_beat
    #    option.per_instance = 1
    #    beat_addr : coverpoint addr {
    #      option.auto_bin_max = 16; }
    #    beat_dir : coverpoint trans_collected.read_write
    #    beat_data : coverpoint data {
    #      option.auto_bin_max = 8; }
    #    beat_wait : coverpoint wait_state {
    #      bins waits[] = { [0:9] }
    #      bins others = { [10:$] }; }
    #    beat_addrXdir : cross beat_addr, beat_dir
    #    beat_addrXdata : cross beat_addr, beat_data
    #  endgroup : cov_trans_beat
    #

    #  // new - constructor
    def __init__(self, name, parent):
        UVMMonitor.__init__(self, name, parent)
        #    cov_trans = new()
        #    cov_trans.set_inst_name({get_full_name(), ".cov_trans"})
        #    cov_trans_beat = new()
        #    cov_trans_beat.set_inst_name({get_full_name(), ".cov_trans_beat"})
        self.trans_collected = ubus_transfer()
        self.item_collected_port = UVMAnalysisPort("item_collected_port", self)
        self.state_port = UVMAnalysisPort("state_port", self)
        self.status = ubus_status("status")

        # The following property is used to store slave address map
        # slave_address_map_info slave_addr_map[string]
        self.slave_addr_map = {}

        self.checks_enable = True
        self.coverage_enable = True
        #  endfunction : new

    #  // set_slave_configs
    def set_slave_configs(self, slave_name, min_addr=0, max_addr=0):
        self.slave_addr_map[slave_name] = slave_address_map_info()
        self.slave_addr_map[slave_name].set_address_map(min_addr, max_addr)
        #  endfunction : set_slave_configs


    def build_phase(self, phase):
        vif = []
        if not (UVMConfigDb.get(self, "", "vif", vif)):
            uvm_fatal("NOVIF", ("virtual interface must be set for: " + self.get_full_name()
                + ".vif"))
        self.vif = vif[0]

    #  // run phase
    
    async def run_phase(self, phase):
        #    fork
        reset_proc = cocotb.fork(self.observe_reset())
        collect_proc = cocotb.fork(self.collect_transactions())
        #    join
        #await [reset_proc, collect_proc.join()]
        await sv.fork_join([reset_proc, collect_proc])
        #  endtask : run_phase
    #

    #  // observe_reset
    
    async def observe_reset(self):
        #    fork
        
        async def rst_start():
            while True:
                await RisingEdge(self.vif.sig_reset)
                self.status.bus_state = RST_START
            self.state_port.write(self.status)

        
        async def rst_stop():
            while True:
                await RisingEdge(self.vif.sig_reset)
                self.status.bus_state = RST_STOP
                self.state_port.write(self.status)

        start_proc = cocotb.fork(rst_start())
        stop_proc = cocotb.fork(rst_stop())

        #await [start_proc, stop_proc.join()]
        await sv.fork_join([start_proc, stop_proc])
        #    join
        #  endtask : observe_reset


    #  // collect_transactions
    
    async def collect_transactions(self):
        while True:
            await self.collect_arbitration_phase()
            await self.collect_address_phase()
            await self.collect_data_phase()
            uvm_info("UBUS_MON", sv.sformatf("Transfer collected :\n%s",
                self.trans_collected.sprint()), UVM_HIGH)
            if (self.checks_enable):
                self.perform_transfer_checks()
            if (self.coverage_enable):
                self.perform_transfer_coverage()
            self.item_collected_port.write(self.trans_collected)
        #  endtask : collect_transactions

    #  // collect_arbitration_phase
    
    async def collect_arbitration_phase(self):
        tmpStr = ""
        # @(posedge vif.sig_clock iff (vif.sig_grant != 0))
        while True:
            await RisingEdge(self.vif.sig_clock)
            if self.vif.sig_grant != 0:
                break
        self.status.bus_state = ARBI
        self.state_port.write(self.status)
        self.begin_tr(self.trans_collected)
        # Check which grant is asserted to determine which master is performing
        # the transfer on the bus.
        #for (int j = 0; j <= 15; j++):
        for j in range(16):
            if (self.vif.sig_grant.value[j] == 1):
                tmpStr = sv.sformatf("masters[%0d]", j)
                self.trans_collected.master = tmpStr
                break
        #  endtask : collect_arbitration_phase
        #

    #  // collect_address_phase
    
    async def collect_address_phase(self):
        await RisingEdge(self.vif.sig_clock)
        self.trans_collected.addr = int(self.vif.sig_addr.value)
        sig_size = int(self.vif.sig_size)
        if sig_size == 0:
            self.trans_collected.size = 1
        elif sig_size == 1:
            self.trans_collected.size = 2
        elif sig_size == 2:
            self.trans_collected.size = 4
        elif sig_size == 3:
            self.trans_collected.size = 8

        self.trans_collected.data = [0] * self.trans_collected.size

        vec = int((int(self.vif.sig_read) << 1) & int(self.vif.sig_write))
        if vec == 0:
            self.trans_collected.read_write = NOP
            self.status.bus_state = NO_OP
            self.state_port.write(self.status)
        elif vec == 1:
            self.trans_collected.read_write = READ
            self.status.bus_state = ADDR_PH
            self.state_port.write(self.status)
        elif vec == 2:
            self.trans_collected.read_write = WRITE
            self.status.bus_state = ADDR_PH
            self.state_port.write(self.status)
        elif vec == 3:
            self.status.bus_state = ADDR_PH_ERROR
            self.state_port.write(self.status)
            if (self.checks_enable):
                uvm_error(self.get_type_name(),
                "Read and Write true at the same time")
        #  endtask : collect_address_phase
        #

    #  // collect_data_phase
    
    async def collect_data_phase(self):
        if (self.trans_collected.read_write != NOP):
            self.check_which_slave()
            for i in range(self.trans_collected.size):
                self.status.bus_state = DATA_PH
                self.state_port.write(self.status)
                while True:
                    await RisingEdge(self.vif.sig_clock)
                    if self.vif.sig_wait == 0:
                        break
                self.trans_collected.data[i] = self.vif.sig_data
            self.num_transactions += 1
            self.end_tr(self.trans_collected)
        else:
            await Timer(0)
        #  endtask : collect_data_phase
        #

    #  // check_which_slave
    def check_which_slave(self):
        slave_name = ""
        slave_found = False

        for slave_name in self.slave_addr_map:
            if (self.slave_addr_map[slave_name].get_min_addr() <= self.trans_collected.addr
              and self.trans_collected.addr <=
              self.slave_addr_map[slave_name].get_max_addr()):
                self.trans_collected.slave = slave_name
                self.slave_found = True
            if self.slave_found is True:
                break

        if slave_found is False:
            uvm_error(self.get_type_name(),
                sv.sformatf("Master attempted a transfer at illegal address 16'h%0h",
                self.trans_collected.addr))
        #  endfunction : check_which_slave
        #

    #  // perform_transfer_checks
    def perform_transfer_checks(self):
        self.check_transfer_size()
        self.check_transfer_data_size()
        #  endfunction : perform_transfer_checks

    #  // check_transfer_size
    def check_transfer_size(self):
        pass
        #   if (trans_collected.read_write != NOP):
        #    assert_transfer_size : assert(trans_collected.size == 1 ||
        #      trans_collected.size == 2 || trans_collected.size == 4 ||
        #      trans_collected.size == 8) else begin
        #      `uvm_error(get_type_name(),
        #        "Invalid transfer size!")
        #    end
        #   end
        #  endfunction : check_transfer_size


    #  // check_transfer_data_size
    def check_transfer_data_size(self):
        pass
        #    if (trans_collected.size != trans_collected.data.size())
        #      `uvm_error(get_type_name(),
        #        "Transfer size field / data size mismatch.")
        #  endfunction : check_transfer_data_size


    #  // perform_transfer_coverage
    def perform_transfer_coverage(self):
        pass
        #    if (trans_collected.read_write != NOP):
        #      -> cov_transaction
        #      for (int unsigned i = 0; i < trans_collected.size; i++):
        #        addr = trans_collected.addr + i
        #        data = trans_collected.data[i]
        #        //wait_state = trans_collected.wait_state[i]
        #        -> cov_transaction_beat
        #      end
        #    end
        #  endfunction : perform_transfer_coverage

    #endclass : ubus_bus_monitor
uvm_component_utils(ubus_bus_monitor)

#  // Provide implementations of virtual methods such as get_type_name and create
#  `uvm_component_utils_begin(ubus_bus_monitor)
#    `uvm_field_int(checks_enable, UVM_DEFAULT)
#    `uvm_field_int(coverage_enable, UVM_DEFAULT)
#    `uvm_field_int(num_transactions, UVM_DEFAULT)
#    `uvm_field_aa_object_string(slave_addr_map, UVM_DEFAULT)
#  `uvm_component_utils_end



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2011 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2010 Cadence Design Systems, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

from uvm.seq.uvm_sequence_item import UVMSequenceItem
from uvm.macros import *
from uvm.reg.uvm_reg_adapter import *
from uvm.reg.uvm_reg_model import *
from uvm.base.uvm_object_globals import *


class apb_rw(UVMSequenceItem):

    READ = 0
    WRITE = 1

    def __init__(self, name="apb_rw"):
        super().__init__(name)
        self.addr = 0  # bit
        self.data = 0  # logic
        self.kind = apb_rw.READ  # kind_e
        self.rand("kind", [apb_rw.READ, apb_rw.WRITE])


    def convert2string(self):
        kind = "READ"
        if self.kind == 1:
            kind = "WRITE"
        return sv.sformatf("kind=%s addr=%0h data=%0h",
                kind, self.addr, self.data)

    #endclass: apb_rw
uvm_object_utils_begin(apb_rw)
uvm_field_int("addr", UVM_ALL_ON | UVM_NOPACK)
uvm_field_int("data", UVM_ALL_ON | UVM_NOPACK)
uvm_object_utils_end(apb_rw)


class reg2apb_adapter(UVMRegAdapter):


    def __init__(self, name="reg2apb_adapter"):
        super().__init__(name)

    def reg2bus(self, rw):
        apb = apb_rw.type_id.create("apb_rw")
        apb.kind = apb_rw.WRITE
        if (rw.kind == UVM_READ):
            apb.kind = apb_rw.READ
        apb.addr = rw.addr
        apb.data = rw.data
        return apb


    def bus2reg(self, bus_item, rw):  # rw must be ref
        apb = None
        arr = []
        if (not sv.cast(arr,bus_item, apb_rw)):
            uvm_fatal("NOT_APB_TYPE", "Provided bus_item is not of the correct type")
            return

        apb = arr[0]
        rw.kind = UVM_WRITE
        if (apb.kind == apb_rw.READ):
            rw.kind = UVM_READ
        rw.addr = apb.addr
        rw.data = apb.data
        rw.status = UVM_IS_OK
        return rw
    #  endfunction
    #
    #endclass
uvm_object_utils(reg2apb_adapter)



#----------------------------------------------------------------------
#   Copyright 2019 Tuomas Poikela (tpoikela)
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
# ----------------------------------------------------------------------

import cocotb
from cocotb.triggers import Timer

from uvm import run_test, UVMTest, UVMComponent
from uvm.base.uvm_root import UVMRoot
from uvm.base.uvm_object_globals import *
from uvm.base.uvm_debug import uvm_debug, UVMDebug
from uvm.macros import (uvm_component_utils, uvm_fatal, uvm_error,
    uvm_info, uvm_component_utils_begin, uvm_component_utils_end,
    uvm_field_string, uvm_field_int)



class DummyTest(UVMTest):
    def __init__(self, name="DummyTest", parent=None):
        super().__init__(name, parent)
        uvm_fatal("DUMM_TEST", "Tests fails if this test is created")


uvm_component_utils(DummyTest)


class CmdLineTest(UVMTest):

    def __init__(self, name="CmdLineTest", parent=None):
        super().__init__(name, parent)
        self.comp = None
        self.comp = MyComponent("my_comp", self)
        self.check_phase_ran = False

    def build_phase(self, phase):
        super().build_phase(phase)

    async def run_phase(self, phase):
        phase.raise_objection(self)
        await Timer(10000, 'ns')
        phase.drop_objection(self)


    def check_phase(self, phase):
        self.check_phase_ran = True
        if self.comp.verb_checked_build_phase is False:
            uvm_fatal("CmdLineTest", "No verb checked in my_comp")


uvm_component_utils(CmdLineTest)


class TestMaxQuitCount(CmdLineTest):

    def __init__(self, name="CmdLineTest", parent=None):
        super().__init__(name, parent)
        self.caught = False

    async def run_phase(self, phase):
        self.comp.set_config_int_test = True
        phase.raise_objection(self)
        await Timer(100, 'ns')
        for i in range(10):
            if i < 9:
                uvm_error("TestMaxQuitCountInfo", "Error " + str(i))
            else:
                # Need to catch this because max_quit_count reached
                try:
                    uvm_error("TestMaxQuitCountInfo", "Last Error " + str(i))
                except Exception as e:
                    uvm_info("TestMaxQuitCount_CAUGHT", "Exception: " + str(e),
                        UVM_NONE)
                    self.caught = True
        await Timer(100, 'ns')
        phase.drop_objection(self)


    def check_phase(self, phase):
        self.check_phase_ran = True
        if self.caught is False:
            uvm_fatal("TestMaxQuitCount TEST FAILED",
                "Did not catch the exception correctly")
        if self.comp.tag != 'matched_correctly':
            uvm_fatal("TestMaxQuitCount TEST FAILED",
                "+uvm_set_config_string did not match correctly")

uvm_component_utils(TestMaxQuitCount)


#------------------------------
# Sub-components for the tests
#------------------------------


class MyComponent(UVMComponent):
    """
    Sub-component created by the tests.
    """

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.verb_checked_build_phase = False
        self.old_rh = self.m_rh
        self.tag = "<not set>"
        self.hex_to_match = 0
        self.oct_to_match = 0
        self.dec_to_match = 0
        self.str_with_ws = "<not set>"
        self.set_config_int_test = False


    def build_phase(self, phase):
        # Need to be called, in order to trigger auto-config
        super().build_phase(phase)
        if self.uvm_report_enabled(UVM_DEBUG):
            uvm_fatal("MyComponent", "Verbosity UVM_DEBUG must not be enabled")

        if not self.uvm_report_enabled(UVM_HIGH):
            uvm_fatal("MyComponent", "Verbosity not UVM_HIGH in build_Phase")
        self.verb_checked_build_phase = True

    
    
    async def run_phase(self, phase):
        self.uvm_report_info("component", "before raising objection", UVM_MEDIUM)
        phase.raise_objection(self)

        if self.uvm_report_enabled(UVM_DEBUG):
            uvm_fatal("MyComponent", "Verbosity UVM_DEBUG must not be enabled")
        #UVMDebug.DEBUG = False
        self.uvm_report_info("component", "after raising objection", UVM_MEDIUM)
        await Timer(1, 'ns')
        uvm_debug(self, 'run_phase', "Objection raised run_phase " +
                self.get_name())
        await Timer(2, 'ns')
        self.uvm_report_info("component", "hello out there!", UVM_MEDIUM)
        await Timer(1005, 'ns')
        assert(self.m_rh is self.old_rh)
        self.uvm_report_info("component", "hello after 1005ns!", UVM_HIGH)
        # verb_level = self.get_report_verbosity_level()

        if not self.uvm_report_enabled(UVM_DEBUG):
            uvm_fatal("MyComponent", "Verbosity UVM_DEBUG must be enabled after 1000ns")
        self.uvm_report_info("component", "finishing up!", UVM_MEDIUM)
        self.uvm_report_fatal("SUPPRESS_FATAL",
            "Should not give FATAL due to cmdline arg +uvm_set_action")
        phase.drop_objection(self)

    def check_phase(self, phase):
        uvm_fatal("FAT2WARN", "Severity should change UVM_FATAL->UVM_WARN")
        if self.set_config_int_test:
            if self.hex_to_match != 0x789:
                uvm_fatal("HEX_NUM_CONFIG_INT_FATAL",
                    "hex_to_match not auto-configured correctly")
            if self.oct_to_match != int('456', 8):
                uvm_fatal("OCT_NUM_CONFIG_INT_FATAL",
                    "oct_to_match not auto-configured correctly")
            if self.dec_to_match < 1000000000:
                uvm_fatal("DEC_NUM_CONFIG_INT_FATAL",
                    "dec_to_match not auto-configured correctly")

uvm_component_utils_begin(MyComponent)
uvm_field_string('tag')
uvm_field_string('str_with_ws')
uvm_field_int('hex_to_match')
uvm_field_int('oct_to_match')
uvm_field_int('dec_to_match')
uvm_component_utils_end(MyComponent)


@cocotb.test()
async def test_cmdline_args(dut):
    await run_test()
    await Timer(1001, 'ns')

    uvm_root = UVMRoot.get()
    test_top = uvm_root.find('uvm_test_top$')
    if test_top.check_phase_ran is False:
        raise Exception('*** TEST FAILED ***\nCheck phase did not run in test_top')



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2011 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2010-2011 Cadence Design Systems, Inc.
#//    Copyright 2019 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//
#

import cocotb
from cocotb.triggers import Timer
from cocotb.clock import Clock

from uvm.base import (uvm_top, UVMCoreService, run_test, UVM_FINISHED, UVM_LOW,
        sv, UVMDebug, UVMConfigDb)
from uvm.comps import UVMTest
from uvm.macros import uvm_fatal, uvm_info
from uvm.reg.sequences import UVMRegHWResetSeq
from uvm.reg.uvm_reg_model import *

from tb_env import tb_env

#UVMDebug.DEBUG = True

class MyClock():

    def __init__(self, dut):
        self.dut = dut
        self.period = 10
        self.unit = 'ns'

    
    async def start(self, n):
        c = Clock(self.dut.clk, self.period, self.unit)
        clk_fork = cocotb.fork(c.start())
        await Timer(n, "NS")
        clk_fork.kill()

class tb_test(UVMTest):
    #
    def __init__(self, name="tb_test", parent=None):
        UVMTest.__init__(self, name, parent)
        self.dut = None

    def build_phase(self, phase):
        UVMTest.build_phase(self, phase)
        self.env = tb_env.type_id.create("env", self)
        UVMConfigDb.set(self.env, "bus", "dut", self.dut)


    #   virtual task run_phase(uvm_phase phase);
    
    async def run_phase(self, phase):

        phase.raise_objection(self)
        env = self.get_child("env")

        if env is None:
            uvm_fatal("test", "Cannot find tb_env")

        await self.reset_dut()
        #env.regmodel.reset()

        clk_fork = cocotb.fork(MyClock(self.dut).start(2000))

        #uvm_reg_sequence seq;
        seq = UVMRegHWResetSeq.type_id.create("reg_hw_rst_seq")
        seq.model = env.regmodel
        print("Before seq.start. Wait for seq state")
        await seq.start(None)
        print("seq.start was called. Wait for seq state")
        await seq.wait_for_sequence_state(UVM_FINISHED)
        print("AFTER yield seq.wait_for_sequence_state(UVM_FINISHED)")

        nwrites = 8
        uvm_info("Test", "Performing " + str(nwrites) + " writes...", UVM_LOW)

        for i in range(nwrites):
            status = []
            uvm_info("WRITE", sv.sformatf("Write[%0d] now", i), UVM_LOW)
            await env.regmodel.user_acp.write(status, sv.random())
            if status[0] == UVM_NOT_OK:
                raise Exception("Reg_write " + i + " failed with UVM_NOT_OK")
            else:
                uvm_info("WRITE_DONE", "UVM_OK status returned for write " +
                    str(i), UVM_LOW)
        status = []
        print("Now mirroring the reg status..")
        await env.regmodel.user_acp.mirror(status, UVM_CHECK)

        await self.reset_dut()
        status = []
        print("Now mirroring the reg status after reset")
        await env.regmodel.user_acp.mirror(status, UVM_CHECK)
        await Timer(100, "NS")

        clk_fork.kill()
        phase.drop_objection(self)
    #endclass

    
    async def reset_dut(self):
        env = self.get_child("env")
        if env is None:
            uvm_fatal("test", "Cannot find tb_env")
        uvm_info("Test", "Resetting DUT...", UVM_LOW)
        self.dut.reset <= 0
        await Timer(5, "NS")
        self.dut.reset <= 1
        env.regmodel.reset()


@cocotb.test()
async def module_top(dut):

    cs_ = UVMCoreService.get()
    test = tb_test("test")
    test.dut = dut

    svr = cs_.get_report_server()
    svr.set_max_quit_count(10)

    await run_test()
    await Timer(1)



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

import cocotb
from cocotb.triggers import Timer

from uvm.comps.uvm_env import UVMEnv
from uvm.macros import uvm_info, uvm_component_utils
from uvm import *
#from packet_pkg import p
from gen_pkg import gen


class env(UVMEnv):
    #    gen gen1
    #

    def __init__(self, name, parent):
        super().__init__(name, parent)
        # use the factory to create the generator
        self.gen1 = gen.type_id.create("gen1", self)


    
    async def run_phase(self, phase):
        phase.raise_objection(self)
        uvm_default_tree_printer.knobs.separator = ""
        for i in range(5):
            await Timer(15, "NS")
            p = self.gen1.get_packet()
            if hasattr(p, 'my_packet_prop') is False:
                uvm_fatal("PKT_ERR", "Wrong packet type created: " + str(p))
            uvm_info("PKTGEN", sv.sformatf("Got packet: %s", p.sprint(uvm_default_tree_printer)), UVM_NONE)
        await Timer(15, "NS")
        phase.drop_objection(self)


#    //Use the macro in a class to implement factory registration along with other
#    //utilities (create, get_type_name). For only factory registration, use the
#    // macro `uvm_component_registry(env,"env").
uvm_component_utils(env)




from .uvm_tlm2_defines import *
from .uvm_tlm2_exports import *
from .uvm_tlm2_generic_payload import *
from .uvm_tlm2_ifs import *
from .uvm_tlm2_imps import *
from .uvm_tlm2_ports import *
from .uvm_tlm2_sockets import *
from .uvm_tlm2_sockets_base import *
from .uvm_tlm_time import *



from os import path
from setuptools import setup, find_packages

version = {}
with open("src/uvm/version.py") as fp:
    exec(fp.read(), version)


def read_file(fname):
    return open(path.join(path.dirname(__file__), fname)).read()


setup(
    name="uvm-python",
    url="https://github.com/tpoikela/uvm-python",
    author="Tuomas Poikela",
    author_email="tuomas.sakari.poikela@gmail.com",
    description=("uvm-python UVM implementation in Python on top of cocotb"),
    long_description=read_file("README.md"),
    long_description_content_type="text/markdown",
    license="Apache 2.0",
    keywords=["UVM", "SystemVerilog", "Verilog", "RTL", "Coverage"],
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    setup_requires=["setuptools_scm",],
    python_requires=">=3.6",
    version=version['__version__'],
    install_requires=[
        #"cocotb @ git+git://github.com/cocotb/cocotb@79792d1b#egg=cocotb",
        "cocotb==1.5.0",
        "cocotb-bus==0.1.1",
        "cocotb-coverage>=1.0.0",
        "regex>=2019.11.1"
    ],
    platforms="any",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)",
    ],
)



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

from uvm.seq import UVMSequence
from uvm.macros.uvm_object_defines import uvm_object_utils
from uvm.macros.uvm_message_defines import uvm_info
from uvm.macros.uvm_sequence_defines import uvm_do_with
from uvm.base import sv, UVM_HIGH
from ubus_transfer import ubus_transfer, READ, WRITE

#//------------------------------------------------------------------------------
#//
#// SEQUENCE: ubus_base_sequence
#//
#//------------------------------------------------------------------------------

#// This sequence raises/drops objections in the pre/post_body so that root
#// sequences raise objections but subsequences do not.

class ubus_base_sequence(UVMSequence):

    def __init__(self, name="ubus_base_seq"):
        UVMSequence.__init__(self, name)
        self.set_automatic_phase_objection(1)
        self.req = ubus_transfer()
        self.rsp = ubus_transfer()

#//------------------------------------------------------------------------------
#//
#// SEQUENCE: read_byte
#//
#//------------------------------------------------------------------------------


class read_byte_seq(ubus_base_sequence):

    def __init__(self, name="read_byte_seq"):
        ubus_base_sequence.__init__(self, name)
        self.start_addr = 0
        self.rand('start_addr', range((1 << 32) - 1))
        self.transmit_delay = 0

    #  constraint transmit_del_ct { (transmit_del <= 10); }


    async def body(self):
        self.req.data = [1234]
        self.req.addr = self.start_addr
        self.req.read_write = READ
        self.req.size = 1
        self.req.error_pos = 1000
        self.req.transmit_delay = self.transmit_delay
        await uvm_do_with(self, self.req, lambda addr: addr == self.start_addr)
        #      { req.addr == start_addr
        #        req.read_write == READ
        #        req.size == 1
        #        req.error_pos == 1000
        #        req.transmit_delay == transmit_del; } )
        rsp = []
        await self.get_response(rsp)
        self.rsp = rsp[0]
        uvm_info(self.get_name(),
            sv.sformatf("%s read : addr = `x{}, data[0] = `x{}",
                self.get_sequence_path(), self.rsp.addr, self.rsp.data[0]),
            UVM_HIGH)


uvm_object_utils(read_byte_seq)


#//------------------------------------------------------------------------------
#//
#// SEQUENCE: read_half_word_seq
#//
#//------------------------------------------------------------------------------
#
#class read_half_word_seq extends ubus_base_sequence
#
#  function new(string name="read_half_word_seq")
#    super.new(name)
#  endfunction
#
#  `uvm_object_utils(read_half_word_seq)
#
#  rand bit [15:0] start_addr
#  rand int unsigned transmit_del = 0
#  constraint transmit_del_ct { (transmit_del <= 10); }
#
#  virtual task body()
#    `uvm_do_with(req,
#      { req.addr == start_addr
#        req.read_write == READ
#        req.size == 2
#        req.error_pos == 1000
#        req.transmit_delay == transmit_del; } )
#    get_response(rsp)
#    `uvm_info(get_type_name(),
#      $sformatf("%s read : addr = `x%0h, data[0] = `x%0h, data[1] = `x%0h",
#      get_sequence_path(), rsp.addr, rsp.data[0], rsp.data[1]), UVM_HIGH)
#  endtask
#
#endclass : read_half_word_seq
#
#
#//------------------------------------------------------------------------------
#//
#// SEQUENCE: read_word_seq
#//
#//------------------------------------------------------------------------------
#
#class read_word_seq extends ubus_base_sequence
#
#  function new(string name="read_word_seq")
#    super.new(name)
#  endfunction
#
#  `uvm_object_utils(read_word_seq)
#
#  rand bit [15:0] start_addr
#  rand int unsigned transmit_del = 0
#  constraint transmit_del_ct { (transmit_del <= 10); }
#
#  virtual task body()
#    `uvm_do_with(req,
#      { req.addr == start_addr
#        req.read_write == READ
#        req.size == 4
#        req.error_pos == 1000
#        req.transmit_delay == transmit_del; } )
#    get_response(rsp)
#    `uvm_info(get_type_name(),
#      $sformatf("%s read : addr = `x%0h, data[0] = `x%0h, \
#      data[1] = `x%0h, data[2] = `x%0h, data[3] = `x%0h",
#      get_sequence_path(), rsp.addr, rsp.data[0], rsp.data[1],
#      rsp.data[2], rsp.data[3]), UVM_HIGH)
#  endtask
#
#endclass : read_word_seq
#
#
#//------------------------------------------------------------------------------
#//
#// SEQUENCE: read_double_word_seq
#//
#//------------------------------------------------------------------------------
#
#class read_double_word_seq extends ubus_base_sequence
#
#  function new(string name="read_double_word_seq")
#    super.new(name)
#  endfunction
#
#  `uvm_object_utils(read_double_word_seq)
#
#  rand bit [15:0] start_addr
#  rand int unsigned transmit_del = 0
#  constraint transmit_del_ct { (transmit_del <= 10); }
#
#  virtual task body()
#    `uvm_do_with(req,
#      { req.addr == start_addr
#        req.read_write == READ
#        req.size == 8
#        req.error_pos == 1000
#        req.transmit_delay == transmit_del; } )
#    get_response(rsp)
#    `uvm_info(get_type_name(),
#      $sformatf("%s read : addr = `x%0h, data[0] = `x%0h, \
#      data[1] = `x%0h, data[2] = `x%0h, data[3] = `x%0h, data[4] = `x%0h, \
#      data[5] = `x%0h, data[6] = `x%0h, data[7] = `x%0h",
#      get_sequence_path(), rsp.addr, rsp.data[0], rsp.data[1], rsp.data[2],
#      rsp.data[3], rsp.data[4], rsp.data[5], rsp.data[6], rsp.data[7]),
#      UVM_HIGH)
#  endtask
#
#endclass : read_double_word_seq
#
#
#//------------------------------------------------------------------------------
#//
#// SEQUENCE: write_byte_seq
#//
#//------------------------------------------------------------------------------

class write_byte_seq(ubus_base_sequence):

    last_data = None

    def __init__(self, name="write_byte_seq"):
        ubus_base_sequence.__init__(self, name)
        self.start_addr = 0
        self.rand('start_addr', range(1 << 16))
        self.data0 = 0
        self.transmit_delay = 0

    #  constraint transmit_del_ct { (transmit_del <= 10); }


    async def body(self):
        req = ubus_transfer()
        req.data.append(self.data0)
        req.addr = self.start_addr
        req.size = 1
        req.error_pos = 1000
        req.read_write = WRITE
        req.transmit_delay = self.transmit_delay
        await uvm_do_with(self, req, lambda addr: addr == self.start_addr,
            lambda read_write: read_write == WRITE,
            lambda data: data[0] == self.data0)
        #      { req.addr == start_addr
        #        req.read_write == WRITE
        #        req.size == 1
        #        req.data[0] == data0
        #        req.error_pos == 1000
        #        req.transmit_delay == transmit_del; } )
        uvm_info(self.get_type_name(),
            sv.sformatf("%s wrote : addr = `x%0h, data[0] = `x%0h",
            self.get_sequence_path(), req.addr, req.data[0]), UVM_HIGH)
        write_byte_seq.last_data = req.data[0]


uvm_object_utils(write_byte_seq)


#//------------------------------------------------------------------------------
#//
#// SEQUENCE: write_half_word_seq
#//
#//------------------------------------------------------------------------------
#
#class write_half_word_seq extends ubus_base_sequence
#
#  function new(string name="write_half_word_seq")
#    super.new(name)
#  endfunction
#
#  `uvm_object_utils(write_half_word_seq)
#
#  rand bit [15:0] start_addr
#  rand bit [7:0] data0
#  rand bit [7:0] data1
#  rand int unsigned transmit_del = 0
#  constraint transmit_del_ct { transmit_del <= 10; }
#
#  virtual task body()
#    `uvm_do_with(req,
#      { req.addr == start_addr;
#        req.read_write == WRITE
#        req.size == 2;
#        req.data[0] == data0; req.data[1] == data1
#        req.error_pos == 1000;
#        req.transmit_delay == transmit_del; } )
#    `uvm_info(get_type_name(),
#      $sformatf("%s wrote : addr = `x%0h, data[0] = `x%0h, data[1] = `x%0h",
#      get_sequence_path(), req.addr, req.data[0], req.data[1]), UVM_HIGH)
#  endtask
#
#endclass : write_half_word_seq
#
#
#//------------------------------------------------------------------------------
#//
#// SEQUENCE: write_word_seq
#//
#//------------------------------------------------------------------------------
#
#class write_word_seq extends ubus_base_sequence
#
#  function new(string name="write_word_seq")
#    super.new(name)
#  endfunction
#
#  `uvm_object_utils(write_word_seq)
#
#  rand bit [15:0] start_addr
#  rand bit [7:0] data0; rand bit [7:0] data1
#  rand bit [7:0] data2; rand bit [7:0] data3
#  rand int unsigned transmit_del = 0
#  constraint transmit_del_ct { (transmit_del <= 10); }
#
#  virtual task body()
#    `uvm_do_with(req,
#      { req.addr == start_addr
#        req.read_write == WRITE
#        req.size == 4
#         req.data[0] == data0; req.data[1] == data1
#         req.data[2] == data2; req.data[3] == data3
#        req.error_pos == 1000
#        req.transmit_delay == transmit_del; } )
#    `uvm_info(get_type_name(),
#      $sformatf("%s wrote : addr = `x%0h, data[0] = `x%0h, \
#      data[1] = `x%0h, data[2] = `x%0h, data[3] = `x%0h",
#      get_sequence_path(), req.addr, req.data[0],
#      req.data[1], req.data[2], req.data[3]),
#      UVM_HIGH)
#  endtask
#
#endclass : write_word_seq
#
#
#//------------------------------------------------------------------------------
#//
#// SEQUENCE: write_double_word_seq
#//
#//------------------------------------------------------------------------------
#
#class write_double_word_seq extends ubus_base_sequence
#
#  function new(string name="write_double_word_seq")
#    super.new(name)
#  endfunction
#
#  `uvm_object_utils(write_double_word_seq)
#
#  rand bit [15:0] start_addr
#  rand bit [7:0] data0; rand bit [7:0] data1
#  rand bit [7:0] data2; rand bit [7:0] data3
#  rand bit [7:0] data4; rand bit [7:0] data5
#  rand bit [7:0] data6; rand bit [7:0] data7
#  rand int unsigned transmit_del = 0
#  constraint transmit_del_ct { (transmit_del <= 10); }
#
#  virtual task body()
#    `uvm_do_with(req,
#      { req.addr == start_addr
#        req.read_write == WRITE
#        req.size == 8
#         req.data[0] == data0; req.data[1] == data1
#         req.data[2] == data2; req.data[3] == data3
#         req.data[4] == data4; req.data[5] == data5
#         req.data[6] == data6; req.data[7] == data7
#        req.error_pos == 1000
#        req.transmit_delay == transmit_del; } )
#    `uvm_info(get_type_name(),
#      $sformatf("Writing  %s : addr = `x%0h, data[0] = `x%0h, \
#      data[1] = `x%0h, data[2] = `x%0h, data[3] = `x%0h, data[4] = `x%0h, \
#      data[5] = `x%0h, data[6] = `x%0h, data[7] = `x%0h",
#      get_sequence_path(), req.addr, req.data[0], req.data[1], req.data[2],
#      req.data[3], req.data[4], req.data[5], req.data[6], req.data[7]),
#      UVM_HIGH)
#  endtask
#
#endclass : write_double_word_seq
#
#



#//----------------------------------------------------------------------
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------
#
#from uvm_pkg import *
#`include "uvm_macros.svh"
#
#`include "usb_xfer.sv"
#`include "host.sv"
#`include "device.sv"
#`include "tb_env.sv"

import cocotb
from tb_env import tb_env
from uvm import run_test

@cocotb.test()
async def tlm2_nb_simple_test(dut):
    env = tb_env("env")
    await run_test()



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

from uvm.base.uvm_object import UVMObject
from uvm.macros import *
from uvm import sv

RANGE_32B = range(0, (1 << 32) - 1)


class packet(UVMObject):
    #    rand int addr
    #    rand int data

    # Use the macro in a class to implement factory registration along with other
    # utilities (create, get_type_name). For only factory registration, use
    # the macro `uvm_object_registry(packet,"packet").

    # Base constraints TODO
    #    constraint c1 { addr inside { [ 0x0: 0x40], [ 0x100: 0x200], [ 0x1000:  0x1fff], [ 0x4000:  0x4fff] }; }
    #    constraint c2 { (addr <=  0x40) -> (data inside { [10:20] } ); }
    #    constraint c3 { (addr >=  0x100  and  addr <=  0x200) -> (data inside { [100:200] } ); }
    #    constraint c4 { (addr >=  0x1000  and  addr <=  0x1fff) -> (data inside { [300:400] } ); }
    #    constraint c5 { (addr >=  0x4000  and  addr <=  0x4fff) -> (data inside { [600:800] } ); }


    # do printing, comparing, etc. These functions can also be automated inside
    # the `uvm_object_utils_begin/end macros if desired. Below show the manual 
    # approach.
    def do_print(self, printer):
        printer.print_field("addr", self.addr, sv.bits(self.addr))
        printer.print_field("data", self.data, sv.bits(self.data))

    def do_compare(self, rhs, comparer):
        rhs_ = None
        if rhs is None:
            return 0
        do_compare = 1
        do_compare &= comparer.compare_field("addr", self.addr, rhs_.addr,
            sv.bits(self.addr))
        do_compare &= comparer.compare_field("data", self.data, rhs_.data,
            sv.bits(self.data))
        return do_compare

    def do_copy(self, rhs):
        rhs_ = rhs
        if rhs is None:
            return
        self.addr = rhs_.addr
        self.data = rhs_.data

    def __init__(self, name="packet"):
        super().__init__(name)
        self.addr = 0
        self.rand('addr', RANGE_32B)
        self.data = 0
        self.rand('data', RANGE_32B)


uvm_object_utils(packet)




# import cocotb
from cocotb.triggers import Timer, Event
from .sv import sv


class UVMScheduler():
    """
    Simple zero-delay/nba scheduler for UVM to hide verilator event details.
    """

    inst = None

    @classmethod
    def get(cls):
        if UVMScheduler.inst is None:
            UVMScheduler.inst = UVMScheduler()
        return UVMScheduler.inst


    def __init__(self):
        self.zero_delay = []
        self.nba_region = []
        self.zero_delay_empty = Event('zero_delay_empty')
        self.enabled = False

    async def run_event_loop(self):
        """
        Each full iteration of the event loop advances simulation by one
        TIMEPRECISION unit.
        """
        self.enabled = True
        while self.enabled:
            while self.has_events() and self.enabled:
                while len(self.zero_delay) > 0:
                    evt, obj = self.zero_delay.pop(0)
                    evt.set()
                while len(self.nba_region) > 0:
                    evt, obj = self.nba_region.pop(0)
                    evt.set()
            await Timer(0)  # Advance simulation

    def has_events(self):
        return (len(self.zero_delay) > 0 or
            len(self.nba_region) > 0)


    async def wait_zero(self, obj):
        evt = Event('evt_0_' + str(sv.urandom()))
        self.zero_delay.append([evt, obj])
        await evt.wait()


    async def wait_nba(self, obj):
        evt = Event('evt_nba_' + str(sv.urandom()))
        self.nba_region.append([evt, obj])
        await evt.wait()


if UVMScheduler.inst is None:
    UVMScheduler.inst = UVMScheduler()



#//
#// -------------------------------------------------------------
#//    Copyright 2011 Synopsys, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

from cocotb.triggers import RisingEdge, FallingEdge
from uvm import *


class vip_driver_cbs(UVMCallback):
    async def pre_tx(self, xactor, char):
        pass
    async def post_tx(self, xactor, char):
        pass

class vip_driver(UVMDriver):


    def __init__(self, name, parent=None):
        super().__init__(name, parent)
        self.m_interrupted = 0
        self.m_proc = None
        self.hier_objection = False
        self.tr = None


    def build_phase(self, phase):
        agent = self.get_parent()
        if agent is not None:
            self.vif = agent.vif.tx
        else:
            vif = []
            if not UVMConfigDb.get(self, "", "vif", vif):
                uvm_fatal("VIP/DRV/NOVIF", "No virtual interface specified for self driver instance")
            self.vif = vif[0]

        self.m_suspend   = 1
        self.m_suspended = 1

    async def pre_reset_phase(self, phase):
        if self.hier_objection:
            phase.raise_objection(self, "Resetting driver")
        self.vif.Tx <= 0
        await self.m_interrupt()
        if self.hier_objection:
            phase.drop_objection(self)

    async def reset_phase(self, phase):
        if self.hier_objection:
            phase.raise_objection(self, "Resetting driver")
        uvm_info("VIP_DRV", "reset_phase started", UVM_LOW)
        await self.reset_and_suspend()
        if self.hier_objection:
            phase.drop_objection(self, "VIP driver has been reset")


    # Abruptly interrupt and suspend this driver
    async def m_interrupt(self):
        self.m_suspend = 1
        if self.m_proc is not None:
            m_proc.kill()
            self.m_proc = None
            self.m_interrupted = 1
            # TODO sv.wait(m_suspended)


    async def reset_and_suspend(self):
        self.vif.Tx <= 0
        await self.m_interrupt()


    async def suspend(self):
        self.m_suspend = 1
        #      wait (m_suspended)
        while self.m_suspended != 1:
            await RisingEdge(self.vif.clk)

    async def resume(self):
        self.m_suspend = 0
        #wait (!m_suspended)
        while self.m_suspended != 0:
            await RisingEdge(self.vif.clk)


    async def run_phase(self, phase):
        self.count = 0

        # self.vif.Tx = 1'bx

        while True:
            if self.tr is not None and not self.m_interrupted:
                self.seq_item_port.item_done()
            self.m_interrupted = 0

            # Reset and suspend
            self.m_suspended = 1
            #wait (!m_suspend)
            while self.m_suspend != 0:
                await RisingEdge(self.vif.clk)
            self.m_suspended = 0
            self.m_proc = cocotb.fork(self.run_phase_fork(phase))
            await sv.fork_join([self.m_proc])

    async def run_phase_fork(self, phase):
        await Timer(2000, "NS")
        while True:

            # Suspend without reset
            if self.m_suspend:
                self.m_suspended = 1
                # wait (!m_suspend)
                while self.m_suspend != 0:
                    await RisingEdge(self.vif.clk)
                self.m_suspended = 0

            if self.count == 0:
                await self.send(0xB2)  # SYNC

            tr = []
            await self.seq_item_port.try_next_item(tr)
            if len(tr) > 0:
                self.tr = tr[0]

            if self.tr is None:
                await self.send(0x81)  # IDLE
            else:
                if self.tr.chr == 0x81 or self.tr.chr == 0xE7:
                    await self.send(0xE7)  # ESC
                    if self.count == 5:
                        await self.send(0xB2)  # SYNC
                        self.count = 0
                    else:
                        self.count += 1
                await self.send(self.tr.chr)

                self.seq_item_port.item_done()
                self.tr = None
            self.count = (self.count + 1) % 6


    async def send(self, data):
        await self.pre_tx(data)
        #      `uvm_do_callbacks(vip_driver, vip_driver_cbs, pre_tx(self, data))
        uvm_info("VIP/DRV/TX", sv.sformatf("Sending 0x%h", data), UVM_HIGH)

        for _ in range(8):
            await FallingEdge(self.vif.clk)
            #vif.Tx = data[7]
            msb = (0x80 & data) >> 7
            self.vif.Tx <= msb
            #data = {data[6:0], data[7]}
            data = (0xFF & (data << 1)) | msb

        #      post_tx(data)
        #      `uvm_do_callbacks(vip_driver, vip_driver_cbs, post_tx(self, data))
        #   endtask: send

    async def pre_tx(self, char):
        pass

    async def post_tx(self, char):
        pass

uvm_component_utils(vip_driver)
uvm_register_cb(vip_driver, vip_driver_cbs)



#
#------------------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2010 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#------------------------------------------------------------------------------
"""
Title: TLM FIFO Classes

This section defines TLM-based FIFO classes.

"""

from .uvm_tlm_fifo_base import UVMTLMFIFOBase
from ..base.uvm_mailbox import UVMMailbox
from ..macros.uvm_message_defines import uvm_error
from .uvm_analysis_port import UVMAnalysisImp


class UVMTLMFIFO(UVMTLMFIFOBase):
    """
    Class: UVMTLMFIFO

    This class provides storage of transactions between two independently running
    processes. Transactions are put into the FIFO via the `UVMTLMFIFOBase.put_export`.
    transactions are fetched from the FIFO in the order they arrived via the
    `get_peek_export`. The `UVMTLMFIFOBase.put_export` and `UVMTLMFIFOBase.get_peek_export`
    are inherited from
    the `UVMTLMFIFOBase` super class, and the interface methods provided by
    these exports are defined by the `TLMIFBaseClass` class.
    """


    type_name = "uvm_tlm_fifo #(T)"

    def __init__(self, name, parent=None, size=1):
        """
        The `name` and `parent` are the normal uvm_component constructor arguments.
        The `parent` should be `null` if the <uvm_tlm_fifo#(T)> is going to be used in a
        statically elaborated construct (e.g., a module). The `size` indicates the
        maximum size of the FIFO; a value of zero indicates no upper bound.

        Args:
            name: Name of the component
            parent: Parent of the component
            size (int): Size of the TLM FIFO.
        """
        UVMTLMFIFOBase.__init__(self, name, parent)
        self.m = UVMMailbox(size)
        self.m_size = size
        self.m_pending_blocked_gets = 0

    def get_type_name(self):
        return UVMTLMFIFO.type_name

    def size(self):
        """
        Returns the capacity of the FIFO, the number of entries
        the FIFO is capable of holding. A return value of 0 indicates the
        FIFO capacity has no limit.

        Returns:
            int: Capacity of the FIFO
        """
        return self.m_size

    def used(self):
        """
          Function: used

          Returns the number of entries put into the FIFO.
        Returns:
            int: Number of entries in the FIFO.
        """
        return self.m.num()

    def is_empty(self):
        """
          Returns 1 when there are no entries in the FIFO, 0 otherwise.
        Returns:
        """
        return self.m.num() == 0

    def is_full(self):
        """
          Returns 1 when the number of entries in the FIFO is equal to its <size>,
          0 otherwise.
        Returns:
        """
        return (self.m_size != 0) and (self.m.num() == self.m_size)


    async def put(self, t):
        await self.m.put(t)
        self.put_ap.write(t)


    async def get(self, t):
        self.m_pending_blocked_gets += 1
        await self.m.get(t)
        self.m_pending_blocked_gets -= 1
        self.get_ap.write(t)


    async def peek(self, t):
        await self.m.peek(t)

    def try_get(self, t):
        if not self.m.try_get(t):
            return False
        self.get_ap.write(t[0])
        return True

    def try_peek(self, t):
        return self.m.try_peek(t)

    def try_put(self, t):
        if not self.m.try_put(t):
            return False
        self.put_ap.write(t)
        return True

    def can_put(self):
        return self.m_size == 0 or self.m.num() < self.m_size

    def can_get(self):
        return self.m.num() > 0 and self.m_pending_blocked_gets == 0

    def can_peek(self):
        return self.m.num() > 0

    def flush(self):
        """
        Removes all entries from the FIFO, after which `used` returns 0
        and `is_empty` returns 1.
        """
        trans = []
        res = 1
        while res:
            res = self.try_get(trans)

        if self.m.num() > 0 and self.m_pending_blocked_gets != 0:
            uvm_error("flush failed",
                "there are blocked gets preventing the flush")



class UVMTLMAnalysisFIFO(UVMTLMFIFO):
    """
    Class: UVMTLMAnalysisFIFO

    An analysis_fifo is a `UVMTLMFIFO` with an unbounded size and a write interface.
    It can be used any place a `UVMAnalysisImp` is used. Typical usage is
    as a buffer between a <uvm_analysis_port> in an initiator component
    and TLM1 target component.
    """

    type_name = "uvm_tlm_analysis_fifo #(T)"


    def __init__(self, name, parent=None):
        """
        This is the standard `UVMComponent` constructor. `name` is the local name
        of this component. The `parent` should be left unspecified when this
        component is instantiated in statically elaborated constructs and must be
        specified when this component is a child of another UVM component.

        Args:
            name (str): Name of the component
            parent (UVMComponent): Parent of the component
        """
        UVMTLMFIFO.__init__(self, name, parent, 0)  # analysis fifo must be unbounded
        #  // Port: analysis_export #(T)
        #  //
        #  // The analysis_export provides the write method to all connected analysis
        #  // ports and parent exports:
        #  //
        #  //|  def write(self, t)
        #  //
        #  // Access via ports bound to this export is the normal mechanism for writing
        #  // to an analysis FIFO.
        #  // See write method of <uvm_tlm_if_base #(T1,T2)> for more information.
        self.analysis_export = UVMAnalysisImp("analysis_export", self)

    def get_type_name(self):
        return UVMTLMAnalysisFIFO.type_name

    def write(self, t):
        self.try_put(t)  # unbounded => must succeed



#//
#//----------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2013      NVIDIA Corporation
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------------

from typing import Any

from .uvm_set_get_dap_base import uvm_set_get_dap_base
from ..base.sv import sv
from ..macros import uvm_error, uvm_object_utils

ERR_MSG1 = ("Attempt to set new value on '%s', but the data access policy " +
    "forbids setting after a get!")

#// Class: uvm_get_to_lock_dap
#// Provides a 'Get-To-Lock' Data Access Policy.
#//
#// The 'Get-To-Lock' Data Access Policy allows for any number of 'sets',
#// until the value is retrieved via a 'get'.  Once 'get' has been called,
#// it is illegal to 'set' a new value.
#//
#// The UVM uses this policy to protect the ~starting phase~ and
#// ~automatic objection~
#// values in <uvm_sequence_base>.
#//
#class uvm_get_to_lock_dap#(type T=int) extends uvm_set_get_dap_base#(T);


class uvm_get_to_lock_dap(uvm_set_get_dap_base):
    #
    # Used for self-references
    # typedef uvm_get_to_lock_dap#(T) this_type;

    type_id = None  # type: Any

    def __init__(self, name="unnamed-uvm_get_to_lock_dap#(T)"):
        uvm_set_get_dap_base.__init__(self, name)
        self.m_locked = False
        self.m_value = None

    #   // Group: Set/Get Interface
    #
    #   // Function: set
    #   // Updates the value stored within the DAP.
    #   //
    #   // ~set~ will result in an error if the value has
    #   // already been retrieved via a call to ~get~.
    def set(self, value):
        if self.m_locked is True:
            uvm_error("UVM/GET_TO_LOCK_DAP/SAG", sv.sformatf(ERR_MSG1, self.get_full_name()))
        else:
            self.m_value = value

    #   // Function: try_set
    #   // Attempts to update the value stored within the DAP.
    #   //
    #   // ~try_set~ will return a 1 if the value was successfully
    #   // updated, or a 0 if the value can not be updated due
    #   // to ~get~ having been called.  No errors will be reported
    #   // if ~try_set~ fails.
    #   virtual function bit try_set(T value);
    #      if (m_locked)
    #        return 0;
    #      else begin
    #         m_value = value;
    #         return 1;
    #      end
    #   endfunction : try_set

    #   // Function: get
    #   // Returns the current value stored within the DAP, and 'locks' the DAP.
    #   //
    #   // After a 'get', the value contained within the DAP cannot
    #   // be changed.
    def get(self):
        self.m_locked = True
        return self.m_value

    #   // Function: try_get
    #   // Retrieves the current value stored within the DAP, and 'locks' the DAP.
    #   //
    #   // ~try_get~ will always return 1.
    #   virtual function bit try_get(output T value);
    #      value = get();
    #      return 1;
    #   endfunction : try_get
    #
    #   // Group: Introspection
    #   //
    #   // The ~uvm_get_to_lock_dap~ cannot support the standard UVM
    #   // instrumentation methods (~copy~, ~clone~, ~pack~ and
    #   // ~unpack~), due to the fact that they would potentially
    #   // violate the access policy.
    #   //
    #   // A call to any of these methods will result in an error.
    #
    #   virtual function void do_copy(uvm_object rhs);
    #      `uvm_error("UVM/GET_TO_LOCK_DAP/CPY",
    #                 "'copy()' is not supported for 'uvm_get_to_lock_dap#(T)'")
    #   endfunction : do_copy
    #
    #   virtual function void do_pack(uvm_packer packer);
    #      `uvm_error("UVM/GET_TO_LOCK_DAP/PCK",
    #                 "'pack()' is not supported for 'uvm_get_to_lock_dap#(T)'")
    #   endfunction : do_pack
    #
    #   virtual function void do_unpack(uvm_packer packer);
    #      `uvm_error("UVM/GET_TO_LOCK_DAP/UPK",
    #                 "'unpack()' is not supported for 'uvm_get_to_lock_dap#(T)'")
    #   endfunction : do_unpack
    #
    #   // Group- Reporting
    #
    #   // Function- convert2string
    #   virtual function string convert2string();
    #      if (m_locked)
    #        return $sformatf("(%s) %0p [LOCKED]", `uvm_typename(m_value), m_value);
    #      else
    #        return $sformatf("(%s) %0p [UNLOCKED]", `uvm_typename(m_value), m_value);
    #   endfunction : convert2string
    #
    #   // Function- do_print
    #   virtual function void do_print(uvm_printer printer);
    #      super.do_print(printer);
    #      printer.print_field_int("lock_state", m_locked, $bits(m_locked));
    #      printer.print_generic("value",
    #                            `uvm_typename(m_value),
    #                            0,
    #                            $sformatf("%0p", m_value));
    #
    #   endfunction : do_print
    #
    #endclass // uvm_get_to_lock_dap


uvm_object_utils(uvm_get_to_lock_dap)



#//
#// -------------------------------------------------------------
#//    Copyright 2011 Synopsys, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------


from cocotb.triggers import RisingEdge, Timer

from uvm import *

from apb.apb_rw import reg2apb_adapter
from apb.apb_agent import apb_agent
from reg_model import reg_dut
from vip import *
from sym_sb import sym_sb
from apb2txrx import apb2txrx

err_msg = ("Environment does not support jumping to phase %s from phase %s. " +
        "Only jumping to \"reset\" is supported")

timeout_ns = 200000 * 10


class rx_isr_seq(UVMRegSequence):

    def __init__(self, name=""):
        super().__init__(name)


    async def pre_body(self):
        self.regmodel = model

    async def body(self):
        status = 0

        # Keep reading data until FIFO is empty
        while not self.regmodel.IntSrc.RxEmpty.get():
            status = []
            await self.regmodel.TxRx.mirror(status)
            status = []
            await self.regmodel.IntSrc.mirror(status)

uvm_object_utils(rx_isr_seq)

TX_ISR = 0
RX_ISR = 1

class tb_env(UVMEnv):

    #   local uvm_status_e   status
    #   local uvm_reg_data_t data
    #   local bit [1:0]      m_isr

    def __init__(self, name, parent=None):
        super().__init__(name, parent)
        self.vif = None
        self.apb = None
        self.reg_dut = None
        self.regmodel = None
        self.tx_src = None
        self.tx_src_seq_port = None
        self.ingress = None
        self.egress = None
        self.adapt = None
        self.m_isr = 0x0
        self.hier_objection = False


    def build_phase(self, phase):
        vif = []
        if not UVMConfigDb.get(self, "", "vif", vif):
           uvm_fatal("TB/ENV/NOVIF", "No virtual interface specified for environment instance")
        self.vif = vif[0]

        self.apb = apb_agent.type_id.create("apb", self)
        if self.regmodel is None:
           self.regmodel = reg_dut.type_id.create("regmodel",None,self.get_full_name())
           self.regmodel.build()
           self.regmodel.lock_model()

        self.tx_src = vip_sequencer.type_id.create("tx_src", self)
        self.tx_src_seq_port = UVMSeqItemPullPort("tx_src_seq_port", self)
        self.vip = vip_agent.type_id.create("vip", self)
        self.vip.hier_objection = self.hier_objection

        self.ingress = sym_sb.type_id.create("ingress", self)
        self.egress = sym_sb.type_id.create("egress", self)
        self.adapt = apb2txrx.type_id.create("adapt", self)

        UVMConfigDb.set(self, "vip.sqr.main_phase", "default_sequence",
                vip_sentence_seq.type_id.get())
        UVMConfigDb.set(self, "tx_src.main_phase","default_sequence", vip_sentence_seq.type_id.get())

        self.m_isr = 0
        self.m_in_shutdown = 0
        self.pull_from_RxFIFO_thread = None

    def has_errors(self):
        return self.ingress.error or self.egress.error


    def connect_phase(self, phase):
        if self.regmodel.get_parent() is None:
            self.reg2apb = reg2apb_adapter()
            self.regmodel.default_map.set_sequencer(self.apb.sqr,self.reg2apb)
            self.regmodel.default_map.set_auto_predict(1)

        self.tx_src_seq_port.connect(self.tx_src.seq_item_export)

        self.apb.mon.ap.connect(self.adapt.apb)

        self.vip.tx_mon.ap.connect(self.ingress.expected)
        self.vip.rx_mon.ap.connect(self.egress.observed)
        self.adapt.tx_ap.connect(self.egress.expected)
        self.adapt.rx_ap.connect(self.ingress.observed)



    # tpoikela: We should not fork anything in phase_started since it's not
    # async. The original SV example abuses this, and runs tasks.
    def phase_started(self, phase):
        name = phase.get_name()
        uvm_info("PHASE_STARTED/TB_ENV", "phase_started(): " + name, UVM_LOW)

        self.m_in_shutdown = 0
        #
        if name == "reset":
            # OK to jump back to reset phase
            if self.pull_from_RxFIFO_thread is not None:
                self.pull_from_RxFIFO_thread.kill()
                self.pull_from_RxFIFO_thread = None

        elif name == "main":
            self.pull_from_RxFIFO_thread = cocotb.fork(self.pull_from_RxFIFO(phase))

        elif name == "shutdown":
            self.m_in_shutdown = 1


    def phase_ended(self, phase):
        goto = phase.get_jump_target()
        name = phase.get_name()
        uvm_info("PHASE_ENDED/TB_ENV", "phase_ended(): " + name, UVM_LOW)

        # This environment supports jump to RESET *only*
        if goto is not None:
           if goto.get_name() != "reset":
               uvm_fatal("ENV/BADJMP", sv.sformatf(err_msg,
                   goto.get_name(), phase.get_name()))


        if phase.get_name() == "main":
            self.m_isr = sv.set_bit(self.m_isr, RX_ISR, 0)

        elif "shutdown":
            self.pull_from_RxFIFO_thread = None
            self.m_in_shutdown = 0


    async def pre_reset_phase(self, phase):
        phase.raise_objection(self, "Waiting for reset to be valid")
        while not (self.vif.rst.value.is_resolvable):
            await RisingEdge(self.vif.clk)
            uvm_info("WAIT_RST", "Waiting reset to become 0/1", UVM_MEDIUM)
        phase.drop_objection(self, "Reset is no longer X")

    async def reset_phase(self, phase):
        phase.raise_objection(self, "Env: Asserting reset for 10 clock cycles")

        uvm_info("TB/TRACE", "Resetting DUT...", UVM_NONE)

        self.vif.rst <= 1
        self.regmodel.reset()
        uvm_info("TB/TRACE", "Waiting VIP reset/suspend", UVM_NONE)
        await self.vip.reset_and_suspend()
        for _ in range(10):
            await RisingEdge(self.vif.clk)
        self.vif.rst <= 0
        uvm_info("TB/TRACE", "Waiting VIP resume", UVM_NONE)
        await self.vip.resume()

        self.m_isr = 0
        self.tx_src.stop_sequences()
        phase.drop_objection(self, "Env: HW reset done")


    async def pre_configure_phase(self, phase):
        phase.raise_objection(self, "Letting the interfaces go idle")
        uvm_info("TB/TRACE", "Configuring DUT...", UVM_NONE)
        for _ in range(10):
            await RisingEdge(self.vif.clk)
        phase.drop_objection(self, "Ready to configure")


    async def configure_phase(self, phase):
        phase.raise_objection(self, "Programming DUT")
        self.regmodel.IntMask.SA.set(1)

        self.regmodel.TxStatus.TxEn.set(1)
        self.regmodel.RxStatus.RxEn.set(1)

        # update the settings BUT without writing the TxRx data register
        status = []
        await self.regmodel.update(status);
        #	begin
        n = [self.regmodel.IntSrc, self.regmodel.IntMask,
                self.regmodel.TxStatus, self.regmodel.RxStatus]
        for reg in n:
            status = []
            await reg.update(status)
        phase.drop_objection(self, "Everything is ready to go")

    async def pre_main_phase_timeout(self):
        for _ in range(100 * 8):
            await RisingEdge(self.vif.sclk)
        uvm_fatal("TB/TIMEOUT", "VIP and/or DUT failed to acquire syncs")

    async def pre_main_phase(self, phase):
        phase.raise_objection(self, "Waiting for VIPs and DUT to acquire SYNC")
        uvm_info("TB/TRACE", "Synchronizing interfaces...", UVM_NONE)

        timeout = cocotb.fork(self.pre_main_phase_timeout())

        # Wait until the VIP has acquired symbol syncs
        while not self.vip.rx_mon.is_in_sync():
            await self.vip.rx_mon.wait_for_sync_change()

        while not self.vip.tx_mon.is_in_sync():
            await self.vip.tx_mon.wait_for_sync_change()

        timeout.kill()

        # Wait until the DUT has acquired symbol sync
        status = []
        await self.regmodel.RxStatus.mirror(status)
        if not self.regmodel.RxStatus.Align.get():
            self.regmodel.IntMask.set(0x000)
            self.regmodel.IntMask.SA.set(0b1)
            status = []
            await self.regmodel.IntMask.update(status)

            #wait (vif.intr)
            while self.vif.intr != 1:
                await RisingEdge(self.vif.clk)

        status = []
        await self.regmodel.IntMask.write(status, 0x000)
        status = []
        await self.regmodel.IntSrc.write(status, -1)
        phase.drop_objection(self, "Everyone is in SYNC")


    #
    # This task is a thread that will span the main and shutdown phase
    #
    async def pull_from_RxFIFO(self, phase):
        shutdown_ph = phase.find_by_name("shutdown")
        shutdown_ph.raise_objection(self, "Pulling data from RxFIFO")

        while True:
            self.m_isr = sv.set_bit(self.m_isr, RX_ISR, 0)
            #
            # When in shutdown, don't wait for the interrupt
            #   wait (vif.intr  or  m_in_shutdown)
            while self.vif.intr == 0 and self.m_in_shutdown == 0:
                await RisingEdge(self.vif.clk)

            #   m_isr[RX_ISR] = 1
            self.m_isr = sv.set_bit(self.m_isr, RX_ISR, 1)

            status = []
            await self.regmodel.IntSrc.mirror(status)
            if self.regmodel.IntSrc.SA.get():
                uvm_fatal("TB/DUT/SYNCLOSS", "DUT has lost SYNC")

            if not self.regmodel.IntSrc.RxHigh.get() and not self.m_in_shutdown:
                self.m_isr = sv.set_bit(self.m_isr, RX_ISR, 0)
                #wait (!m_isr)
                while self.m_isr == 1:
                    await RisingEdge(self.vif.clk)
                continue

            rx_seq = rx_isr_seq.type_id.create("rx_seq",None, self.get_full_name())
            rx_seq.model = self.regmodel
            await rx_seq.start(None)

            #   m_isr[RX_ISR] = 0
            self.m_isr = sv.set_bit(self.m_isr, RX_ISR, 0)

            if self.m_in_shutdown:
                shutdown_ph.drop_objection(self, "All data pulled from RxFIFO")
                break



    async def main_phase(self, phase):
        uvm_info("TB/TRACE", "Applying primary stimulus...", UVM_NONE)
        timeout_proc = cocotb.fork(self.timeout_and_finish(phase))
        main_proc = cocotb.fork(self.run_main_proc(phase))
        await sv.fork_join([timeout_proc, main_proc])

    async def timeout_and_finish(self, phase):
        await Timer(timeout_ns, "NS")
        obj = phase.get_objection()
        obj.display_objections()
        uvm_fatal("ERR/TIMEOUT", "$finish. Timeout reached")
        # $finish

    async def run_main_proc(self, phase):
        status = []
        ph_obj = phase.get_objection()
        phase.raise_objection(self, "Configuring ISR")
        self.regmodel.IntMask.set(0)
        self.regmodel.IntMask.SA.set(1)
        self.regmodel.IntMask.RxHigh.set(1)
        self.regmodel.IntMask.TxLow.set(1)
        await self.regmodel.IntMask.update(status)

        while True:
            phase.drop_objection(self, "ISR ready DUT-> stimulus")
            self.m_isr = sv.set_bit(self.m_isr, TX_ISR, 0)
            #   m_isr[TX_ISR] = 0

            uvm_info("TB_ENV", "Waiting vif.intr now", UVM_MEDIUM)
            while not self.vif.intr:
                await RisingEdge(self.vif.intr)
                #   wait (vif.intr)

            #   m_isr[TX_ISR] = 1
            self.m_isr = sv.set_bit(self.m_isr, TX_ISR, 1)
            phase.raise_objection(self, "Applying DUT-> stimulus")

            status = []
            await self.regmodel.IntSrc.mirror(status)
            if not self.regmodel.IntSrc.TxLow.get():
                self.m_isr = sv.set_bit(self.m_isr, TX_ISR, 1)
                # m_isr[TX_ISR] = 0
                while self.m_isr == 0x0:
                    #wait (!m_isr)
                    await RisingEdge(self.vif.clk)
                continue

            #   m_isr[TX_ISR] = 0
            self.m_isr = sv.set_bit(self.m_isr, TX_ISR, 0)

            self.regmodel.IntMask.TxLow.set(0)
            status = []
            await self.regmodel.IntMask.update(status)

            # Stop supplying data once it is full
            # or the egress scoreboard has had enough
            while not self.regmodel.IntSrc.TxFull.get() > 0:
                phase.drop_objection(self, "Waiting for DUT-> stimulus from tx_src sequencer")
                tr = []
                await self.tx_src_seq_port.get_next_item(tr)
                self.tx_src_seq_port.item_done()
                phase.raise_objection(self, "Applying DUT-> stimulus from tx_src sequencer")
                await self.regmodel.TxRx.write(status, tr[0].chr)

                await self.regmodel.IntSrc.mirror(status)

            self.regmodel.IntMask.TxLow.set(1)
            await self.regmodel.IntMask.update(status)


    async def shutdown_phase(self, phase):
        phase.raise_objection(self, "Draining the DUT")

        uvm_info("TB/TRACE", "Draining the DUT...", UVM_NONE)

        if not regmodel.IntSrc.TxEmpty.get():
            # Wait for TxFIFO to be empty
            status = []
            self.regmodel.IntMask.write(status, 0x001)

            #wait (vif.intr)
            while vif.intr != 1:
                await RisingEdge(self.vif.clk)

        # Make sure the last symbol is transmitted
        # repeat (16) @(posedge vif.sclk)
        for _ in range(16):
            await RisingEdge(self.vif.clk)

        phase.drop_objection(self, "DUT is empty")


    def report_phase(self, phase):
        cs_ = UVMCoreService.get()
        svr = cs_.get_report_server()

        if (svr.get_severity_count(UVM_FATAL) +
                svr.get_severity_count(UVM_ERROR) == 0):
            print("** UVM TEST PASSED **\n")
        else:
            print("!! UVM TEST FAILED !!\n")

uvm_component_utils(tb_env)



#//----------------------------------------------------------------------
#//   Copyright 2010-2011 Synopsys, Inc
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

import cocotb

from uvm import (UVMComponent, sv, wait)
from uvm.macros import uvm_info, uvm_component_utils
from uvm.tlm2 import *

from usb_xfer import (usb_xfer, USB_TLM_TOKEN,
    USB_TLM_DATA,  USB_TLM_HANDSHAKE)


class host(UVMComponent):


    def __init__(self, name="host", parent=None):
        super().__init__(name, parent)
        #   uvm_tlm_nb_initiator_socket#(host, usb_xfer, usb_tlm_phase) sock
        self.sock = UVMTLMNbInitiatorSocket("sock", self)
        self.xfer = None  # usb_xfer
        self.ph = None  # usb_tlm_phase
        self.delay = UVMTLMTime()
        self.sync = 0  # uvm_tlm_sync_e

    #   // Backward path
    def nb_transport_bw(self, xfer, ph, delay):
        uvm_info("USB/HST/BWD", sv.sformatf("%s @%0d: %s",
            self.ph.name(), self.delay.get_realtime(1, "ns"),
            xfer.convert2string()), UVM_LOW)

        assert(xfer == self.xfer)
        self.ph    = ph
        self.delay = delay

        if xfer.kind == usb_xfer.OUT:
            assert(ph == USB_TLM_HANDSHAKE)
            self.sync = uvm_tlm_sync_e.UVM_TLM_COMPLETED
            return uvm_tlm_sync_e.UVM_TLM_COMPLETED
        elif xfer.kind == usb_xfer.IN:
            assert(ph == USB_TLM_DATA)
            # Could do an early completion too
            return uvm_tlm_sync_e.UVM_TLM_ACCEPTED
        return uvm_tlm_sync_e.UVM_TLM_COMPLETED


    #   //
    #   // Execute an OUT followed by an IN bulk transfer
    #   //
    async def run_phase(self, phase):
        phase.raise_objection(self)
        self.xfer = usb_xfer.type_id.create("xfer",None, self.get_full_name())

        # OUT bulk transfer

        done = False
        while not done:
            self.xfer.kind = usb_xfer.OUT
            self.xfer.addr = 0x7A
            self.xfer.endp = 0x5

            self.ph = USB_TLM_TOKEN
            self.sync = await self.sock.nb_transport_fw(self.xfer, self.ph, self.delay)
            print("self.sync is " + self.sync.name)
            if self.sync == uvm_tlm_sync_e.UVM_TLM_COMPLETED:
                uvm_info("USB/HST/OUT/REFUSED", "Device refused the transfer",
                    UVM_LOW)
                break

            self.ph = USB_TLM_DATA
            tdata = [0xDE, 0xAD, 0xBE, 0xEF]
            self.xfer.data = tdata
            self.sync = await self.sock.nb_transport_fw(self.xfer, self.ph, self.delay)
            if self.sync == uvm_tlm_sync_e.UVM_TLM_COMPLETED:
                uvm_info("USB/HST/OUT/EARLY", "Device completed OUT transfer early",
                         UVM_LOW)
                break

            # Wait for the device to reply
            # TODO
            # wait(self.sync == uvm_tlm_sync_e.UVM_TLM_COMPLETED)
            uvm_info("USB/HST/OUT/NORMAL", "OUT Transfer completed normally",
                UVM_LOW)
            done = True


        uvm_info("USB/HST/OUT/DONE", self.xfer.convert2string(), UVM_NONE)

        # IN bulk transfer
        done = False
        while not done:
            # Ok to reuse the same XFER instance
            self.xfer.kind = usb_xfer.IN
            self.xfer.addr = 0x7A
            self.xfer.endp = 0x5
            self.xfer.data = []

            self.ph = USB_TLM_TOKEN
            self.sync = await self.sock.nb_transport_fw(self.xfer, self.ph, self.delay)
            if self.sync == uvm_tlm_sync_e.UVM_TLM_COMPLETED:
                uvm_info("USB/HST/IN/REFUSED", "Device refused the transfer",
                        UVM_LOW)
                break
            if self.sync == uvm_tlm_sync_e.UVM_TLM_UPDATED:
                uvm_info("USB/HST/IN/EARLY", "Device returned bulk data early",
                        UVM_LOW)
            else:
                # Wait for the device to reply
                # TODO wait(ph == USB_TLM_DATA)
                uvm_info("USB/HST/IN/NORMAL", "IN data returned normally",
                        UVM_LOW)

            self.ph = USB_TLM_HANDSHAKE
            self.xfer.status = usb_xfer.ACK
            self.sync = await self.sock.nb_transport_fw(self.xfer, self.ph, self.delay)
            assert(self.sync == uvm_tlm_sync_e.UVM_TLM_COMPLETED)
            done = True

        uvm_info("USB/HST/IN/DONE", self.xfer.convert2string(), UVM_NONE)
        phase.drop_objection(self)


uvm_component_utils(host)



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

from uvm.base import UVMTransaction
from uvm.macros import uvm_object_utils

class packet(UVMTransaction):

    #`ifndef NO_RAND
    #  rand
    #`endif
    #  int addr;

    #  constraint c { addr >= 0 && addr < 'h100; }

    def __init__(self, name):
        UVMTransaction.__init__(self, name)
        self.addr = 0

uvm_object_utils(packet)
#uvm_field_int(addr, UVM_ALL_ON)



#//------------------------------------------------------------------------------
#//   Copyright 2014 Synopsys, Inc.
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//------------------------------------------------------------------------------

#//
#// Title: Global Macros
#//------------------------
#// Group: Global object Macro definitions can be used in multiple locations
#//------------------------
#//
#// MACRO: `UVM_MAX_STREAMBITS
#//
#// Defines the maximum bit vector size for integral types.
#// Used to set uvm_bitstream_t
#
#`ifndef UVM_MAX_STREAMBITS
# `define UVM_MAX_STREAMBITS 4096
#`endif
#
#
#// MACRO: `UVM_PACKER_MAX_BYTES
#//
#// Defines the maximum bytes to allocate for packing an object using
#// the <uvm_packer>. Default is <`UVM_MAX_STREAMBITS>, in ~bytes~.
#
#`ifndef UVM_PACKER_MAX_BYTES
# `define UVM_PACKER_MAX_BYTES `UVM_MAX_STREAMBITS
#`endif

#//------------------------
#// Group: Global Time Variable definitions that can be used in multiple locations
#//------------------------

#// variable: UVM_DEFAULT_TIMEOUT
#//
#// The default timeout for simulation, if not overridden by
#// `UVMRoot.set_timeout` or <uvm_cmdline_processor::+UVM_TIMEOUT>
#//

UVM_DEFAULT_TIMEOUT = 9200
UVM_DEFAULT_TIMEOUT_UNIT = "s"



#//
#//------------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2013 NVIDIA Corporation
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//------------------------------------------------------------------------------

from .uvm_object_globals import *
from .uvm_scope_stack import UVMScopeStack
from .sv import sv
from ..macros.uvm_message_defines import uvm_error, uvm_warning
from typing import List


SIZEOF_INT = 32
MASK_INT = 0xFFFFFFFF


class UVMPacker(object):
    """
    The UVMPacker class provides a policy object for packing and unpacking
    uvm_objects. The policies determine how packing and unpacking should be done.
    Packing an object causes the object to be placed into a bit (byte or int)
    array. If the `uvm_field_* macro are used to implement pack and unpack,
    by default no metadata information is stored for the packing of dynamic
    objects (strings, arrays, class objects).
    """


    bitstream = []  # local bits for (un)pack_bytes
    fabitstream = []  # field automation bits for (un)pack_bytes

    #  //----------------//
    #  // Group: Packing //
    #  //----------------//

    def __init__(self):
        #  //------------------//
        #  // Group: Variables //
        #  //------------------//

        #  // Variable: physical
        #  //
        #  // This bit provides a filtering mechanism for fields.
        #  //
        #  // The <abstract> and physical settings allow an object to distinguish between
        #  // two different classes of fields. It is up to you, in the
        #  // `UVMObject.do_pack` and `UVMObject.do_unpack` methods, to test the
        #  // setting of this field if you want to use it as a filter.
        #  bit physical = 1
        self.physical = 1


        #  // Variable: abstract
        #  //
        #  // This bit provides a filtering mechanism for fields.
        #  //
        #  // The abstract and physical settings allow an object to distinguish between
        #  // two different classes of fields. It is up to you, in the
        #  // `UVMObject.do_pack` and `UVMObject.do_unpack` routines, to test the
        #  // setting of this field if you want to use it as a filter.
        #  bit abstract
        self.abstract = 0


        #  // Variable: use_metadata
        #  //
        #  // This flag indicates whether to encode metadata when packing dynamic data,
        #  // or to decode metadata when unpacking.  Implementations of `UVMObject.do_pack`
        #  // and `UVMObject.do_unpack` should regard this bit when performing their
        #  // respective operation. When set, metadata should be encoded as follows:
        #  //
        #  // - For strings, pack an additional ~null~ byte after the string is packed.
        #  //
        #  // - For objects, pack 4 bits prior to packing the object itself. Use 4'b0000
        #  //   to indicate the object being packed is ~null~, otherwise pack 4'b0001 (the
        #  //   remaining 3 bits are reserved).
        #  //
        #  // - For queues, dynamic arrays, and associative arrays, pack 32 bits
        #  //   indicating the size of the array prior to packing individual elements.
        #  bit use_metadata
        self.use_metadata = 0


        #  // Variable: big_endian
        #  //
        #  // This bit determines the order that integral data is packed (using
        #  // <pack_field>, <pack_field_int>, <pack_time>, or <pack_real>) and how the
        #  // data is unpacked from the pack array (using <unpack_field>,
        #  // <unpack_field_int>, <unpack_time>, or <unpack_real>). When the bit is set,
        #  // data is associated msb to lsb; otherwise, it is associated lsb to msb.
        #  //
        #  // The following code illustrates how data can be associated msb to lsb and
        #  // lsb to msb:
        #  //
        #  //|  class mydata extends uvm_object;
        #  //|
        #  //|    logic[15:0] value = 'h1234;
        #  //|
        #  //|    function void do_pack (UVMPacker packer);
        #  //|      packer.pack_field_int(value, 16);
        #  //|    endfunction
        #  //|
        #  //|    function void do_unpack (UVMPacker packer);
        #  //|      value = packer.unpack_field_int(16);
        #  //|    endfunction
        #  //|  endclass
        #  //|
        #  //|  mydata d = new;
        #  //|  bit bits[];
        #  //|
        #  //|  initial begin
        #  //|    d.pack(bits);  // 'b0001001000110100
        #  //|    uvm_default_packer.big_endian = 0;
        #  //|    d.pack(bits);  // 'b0010110001001000
        #  //|  end
        #  bit big_endian = 1
        self.big_endian = 1

        # variables and methods primarily for internal use
        self.count = 0                # used to count the number of packed bits
        self.scope = UVMScopeStack()
        self.reverse_order = 0      # flip the bit order around
        self.byte_size     = 8   # set up bytesize for endianess
        self.word_size     = 16  # set up worksize for endianess
        self.nopack = 0          # only count packable bits
        self.policy = UVM_DEFAULT_POLICY
        self.m_bits = 0x0  # uvm_pack_bitstream_t
        self.m_packed_size = 0


    #  // Function: pack_field
    #  //
    #  // Packs an integral value (less than or equal to 4096 bits) into the
    #  // packed array. ~size~ is the number of bits of ~value~ to pack.
    #
    #  extern def pack_field(self,uvm_bitstream_t value, int size):
    def pack_field(self, value, size) -> None:
        #  for (int i=0; i<size; i++)
        if self.big_endian == 1:
            flipped = self.flip_bit_order(value, size)
            self.m_bits |= flipped << self.count
            # self.m_bits[count+i] = value[size-1-i]
        else:
            self.m_bits |= value << self.count
        self.count += size


    #  // Function: pack_field_int
    #  //
    #  // Packs the integral value (less than or equal to 64 bits) into the
    #  // pack array.  The ~size~ is the number of bits to pack, usually obtained by
    #  // ~$bits~. This optimized version of <pack_field> is useful for sizes up
    #  // to 64 bits.
    def pack_field_int(self, value: int, size: int) -> None:
        if self.big_endian == 1:
            flipped = self.flip_bit_order(value, size)
            self.m_bits |= flipped << self.count
            #self.m_bits[self.count+i] = value[size-1-i]
        else:
            self.m_bits |= value << self.count
            #self.m_bits[self.count+i] = value[i]
        self.count += size


    #  // Function: pack_bits
    #  //
    #  // Packs bits from upacked array of bits into the pack array.
    #  //
    #  // See <pack_ints> for additional information.
    #  extern def pack_bits(self,ref bit value[], input int size = -1):


    #  // Function: pack_bytes
    #  //
    #  // Packs bits from an upacked array of bytes into the pack array.
    #  //
    #  // See <pack_ints> for additional information.
    #  extern def pack_bytes(self,ref byte value[], input int size = -1):
    def pack_bytes(self, value, size=-1) -> None:
        max_size = len(value) * 8

        if size < 0:
            size = max_size

        if size > max_size:
            uvm_error("UVM/BASE/PACKER/BAD_SIZE",
            sv.sformatf("pack_bytes called with size '%0d', which exceeds value size of '%0d'",
                size, max_size))
            return
        else:
            for i in range(len(value)):
                byte = value[i]
                if self.big_endian == 1:
                    byte = self.flip_bit_order(value[len(value)-1-i], 8)
                self.m_bits |= byte << self.count
                self.count += 8


    #  // Function: pack_ints
    #  //
    #  // Packs bits from an unpacked array of ints into the pack array.
    #  //
    #  // The bits are appended to the internal pack array.
    #  // This method allows for fields of arbitrary length to be
    #  // passed in, using the SystemVerilog ~stream~ operator.
    #  //
    #  // For example
    #  // | bit[511:0] my_field;
    #  // | begin
    #  // |   int my_stream[];
    #  // |   { << int {my_stream}} = my_field;
    #  // |   packer.pack_ints(my_stream);
    #  // | end
    #  //
    #  // When appending the stream to the internal pack array, the packer will obey
    #  // the value of <big_endian> (appending the array from MSB to LSB if set).
    #  //
    #  // An optional ~size~ parameter is provided, which defaults to '-1'.  If set
    #  // to any value greater than '-1' (including 0), then the packer will use
    #  // the size as the number of bits to pack, otherwise the packer will simply
    #  // pack the entire stream.
    #  //
    #  // An error will be asserted if the ~size~ has been specified, and exceeds the
    #  // size of the source array.
    #  //
    #  extern def pack_ints(self,ref int value[], input int size = -1):
    def pack_ints(self, value, size=-1) -> None:
        max_size = len(value) * SIZEOF_INT

        if size < 0:
            size = max_size

        if size > max_size:
            uvm_error("UVM/BASE/PACKER/BAD_SIZE",
                sv.sformatf("pack_ints called with size '%0d', which exceeds value size of '%0d'",
                    size, max_size))
            return
        else:
            for i in range(len(value)):
                int_num = value[i]
                if self.big_endian == 1:
                    int_num = self.flip_bit_order(value[len(value)-1-i],
                        SIZEOF_INT)
                self.m_bits |= int_num << self.count
                self.count += SIZEOF_INT


    #  // Function: pack_string
    #  //
    #  // Packs a string value into the pack array.
    #  //
    #  // When the metadata flag is set, the packed string is terminated by a ~null~
    #  // character to mark the end of the string.
    #  //
    #  // This is useful for mixed language communication where unpacking may occur
    #  // outside of SystemVerilog UVM.
    #
    #  extern def pack_string(self,string value):
    def pack_string(self, value) -> None:
        bytearr = value.encode()

        size = 8 * len(bytearr)
        bits = int(bytearr.hex(), 16)
        if self.big_endian == 1:
            bits = self.flip_bit_order(bits, -1)
        self.m_bits |= bits << self.count
        self.count += size
        if self.use_metadata == 1:
            pass
            # TODO self.m_bits |= 0 << self.count
        #  byte b
        #  foreach (value[index]):
        #    if(self.big_endian == 0)
        #      self.m_bits[count +: 8] = value[index]
        #    else begin
        #      b = value[index]
        #      for(int i=0; i<8; ++i)
        #        self.m_bits[count+i] = b[7-i]
        #    end
        #    count += 8
        #  end
        #  if(use_metadata == 1):
        #    self.m_bits[count +: 8] = 0
        #    count += 8
        #  end
        #endfunction


    #  // Function: pack_time
    #  //
    #  // Packs a time ~value~ as 64 bits into the pack array.
    #
    #  extern def pack_time(self,time value):;


    #  // Function: pack_real
    #  //
    #  // Packs a real ~value~ as 64 bits into the pack array.
    #  //
    #  // The real ~value~ is converted to a 6-bit scalar value using the function
    #  // $real2bits before it is packed into the array.
    #
    #  extern def pack_real(self,real value):


    #  // Function: pack_object
    #  //
    #  // Packs an object value into the pack array.
    #  //
    #  // A 4-bit header is inserted ahead of the string to indicate the number of
    #  // bits that was packed. If a ~null~ object was packed, then this header will
    #  // be 0.
    #  //
    #  // This is useful for mixed-language communication where unpacking may occur
    #  // outside of UVM.
    #
    #  extern def pack_object(self,uvm_object value):
    def pack_object(self, value) -> None:
        if value in value._m_uvm_status_container.cycle_check:
            uvm_warning("CYCFND", sv.sformatf("Cycle detected for object @%0d during pack",
                value.get_inst_id()))
            return

        value._m_uvm_status_container.cycle_check[value] = 1

        if((self.policy != UVM_REFERENCE) and value is not None):
            if self.use_metadata == 1:
                pass
                #self.m_bits[count +: 4] = 1
                #count += 4; // to better debug when display packed bits in hexadecimal

            self.scope.down(value.get_name())
            value._m_uvm_field_automation(None, UVM_PACK,"")
            value.do_pack(self)
            self.scope.up()
        elif self.use_metadata == 1:
            pass
            #self.m_bits[count +: 4] = 0
            #count += 4
        del value._m_uvm_status_container.cycle_check[value]


    #  //------------------//
    #  // Group: Unpacking //
    #  //------------------//

    #  // Function: is_null
    #  //
    #  // This method is used during unpack operations to peek at the next 4-bit
    #  // chunk of the pack data and determine if it is 0.
    #  //
    #  // If the next four bits are all 0, then the return value is a 1; otherwise
    #  // it is 0.
    #  //
    #  // This is useful when unpacking objects, to decide whether a new object
    #  // needs to be allocated or not.
    #
    #  extern def is_null(self):


    #  // Function: unpack_field
    #  //
    #  // Unpacks bits from the pack array and returns the bit-stream that was
    #  // unpacked. ~size~ is the number of bits to unpack; the maximum is 4096 bits.
    #
    #  extern def unpack_field(self,int size):
    def unpack_field(self, size):
        return self.unpack_field_int(size)
        #  unpack_field =  0b0
        #  if (self.enough_bits(size,"integral")):
        #    count += size
        #    for (int i=0; i<size; i++)
        #      if(self.big_endian == 1)
        #        unpack_field[i] = self.m_bits[count-i-1]
        #      else
        #        unpack_field[i] = self.m_bits[count-size+i]
        #  end
        #endfunction


    #  // Function: unpack_field_int
    #  //
    #  // Unpacks bits from the pack array and returns the bit-stream that was
    #  // unpacked.
    #  //
    #  // ~size~ is the number of bits to unpack; the maximum is 64 bits.
    #  // This is a more efficient variant than unpack_field when unpacking into
    #  // smaller vectors.
    #
    #  extern def unpack_field_int(self,int size):
    def unpack_field_int(self, size) -> int:
        unpack_field_int = 0x0
        count_before = self.count
        if self.enough_bits(size,"integral"):
            self.count += size
            for i in range(size):
                if self.big_endian:
                    bit_sel = self.count-i-1
                    bit_sel = (1 << bit_sel)
                    unpack_field_int |= self.m_bits & bit_sel
                else:
                    bit_sel = self.count-size+i
                    bit_sel = (1 << bit_sel)
                    unpack_field_int |= self.m_bits & bit_sel
        unpack_field_int >>= count_before
        if self.big_endian:
            unpack_field_int = self.flip_bit_order(unpack_field_int, size)
        return unpack_field_int


    #  // Function: unpack_bits
    #  //
    #  // Unpacks bits from the pack array into an unpacked array of bits.
    #  //
    #  extern def unpack_bits(self,ref bit value[], input int size = -1):
    #

    #  // Function: unpack_bytes
    #  //
    #  // Unpacks bits from the pack array into an unpacked array of bytes.
    #  //
    #  extern def unpack_bytes(self,ref byte value[], input int size = -1):
    def unpack_bytes(self, value, size=-1):
        max_size = len(value) * 8
        if size < 0:
            size = max_size

        if size > max_size:
            uvm_error("UVM/BASE/PACKER/BAD_SIZE",
                sv.sformatf("unpack_bytes called with size '%0d', which exceeds value size of '%0d'",
                    size, len(value)))
            return []
        else:
            if self.enough_bits(size, "integral"):
                self.count += size
                for b in range(len(value)):
                    byte = (self.m_bits >> b * 8) & 0xFF
                    if self.big_endian == 1:
                        byte = self.flip_bit_order(byte, 8)
                        value[len(value)-1-b] = byte
                    else:
                        value[b] = byte
            return value


    #  // Function: unpack_ints
    #  //
    #  // Unpacks bits from the pack array into an unpacked array of ints.
    #  //
    #  // The unpacked array is unpacked from the internal pack array.
    #  // This method allows for fields of arbitrary length to be
    #  // passed in without expanding into a pre-defined integral type first.
    #  //
    #  // For example
    #  // | bit[511:0] my_field;
    #  // | begin
    #  // |   int my_stream[] = new[16]; // 512/32 = 16
    #  // |   packer.unpack_ints(my_stream);
    #  // |   my_field = {<<{my_stream}};
    #  // | end
    #  //
    #  // When unpacking the stream from the internal pack array, the packer will obey
    #  // the value of <big_endian> (unpacking the array from MSB to LSB if set).
    #  //
    #  // An optional ~size~ parameter is provided, which defaults to '-1'.  If set
    #  // to any value greater than '-1' (including 0), then the packer will use
    #  // the size as the number of bits to unpack, otherwise the packer will simply
    #  // unpack the entire stream.
    #  //
    #  // An error will be asserted if the ~size~ has been specified, and
    #  // exceeds the size of the target array.
    #  //
    #  extern def unpack_ints(self,ref int value[], input int size = -1):
    def unpack_ints(self, value, size=-1):
        max_size = len(value) * SIZEOF_INT
        if size < 0:
            size = max_size

        if size > max_size:
            uvm_error("UVM/BASE/PACKER/BAD_SIZE",
                sv.sformatf("unpack_ints called with size '%0d', which exceeds value size of '%0d'",
                   size, len(value)))
            return
        else:
            if self.enough_bits(size, "integral"):
                self.count += size
                for i in range(len(value)):
                    int_num = (self.m_bits >> i * SIZEOF_INT) & 0xFFFFFFFF
                    if self.big_endian == 1:
                        int_num = self.flip_bit_order(int_num, SIZEOF_INT)
                        value[len(value)-1-i] = int_num
                    else:
                        value[i] = int_num
            return value


    #  // Function: unpack_string
    #  //
    #  // Unpacks a string.
    #  //
    #  // num_chars bytes are unpacked into a string. If num_chars is -1 then
    #  // unpacking stops on at the first ~null~ character that is encountered.
    #// If num_chars is not -1, then the user only wants to unpack a
    #// specific number of bytes into the string.
    def unpack_string(self, num_chars=-1):
        #  byte b
        i = 0
        is_null_term = 0  # Assumes a ~None~ terminated string
        #  int i; i=0
        if num_chars == -1:
            is_null_term = 1

        #val_to_decode = 0x0
        # We'll use bytearray to decode this, so need to find the num of bytes

        byte_arr = bytearray()
        #unpack_string = 0
        curr_byte = (self.m_bits >> self.count) & 0xFF
        while (self.enough_bits(8,"string", is_error=False) and
              ((curr_byte != 0) or (is_null_term == 0)) and
              ((i < num_chars) or (is_null_term == 1))):
            # silly, because cannot append byte/char to string
            #unpack_string = unpack_string + " "
            #if self.big_endian == 0:
            #    unpack_string[i] = self.m_bits[count +: 8]
            #else:
            #    for(int j=0; j<8; ++j)
            #        b[7-j] = self.m_bits[count+j]
            #  unpack_string[i] = b
            i += 1
            byte_arr.insert(0, curr_byte)
            self.count += 8
            curr_byte = (self.m_bits >> self.count) & 0xFF
        if self.enough_bits(8,"string", is_error=False):
            self.count += 8
        return byte_arr.decode()
        #return unpack_string


    #  // Function: unpack_time
    #  //
    #  // Unpacks the next 64 bits of the pack array and places them into a
    #  // time variable.
    #
    #  extern def unpack_time(self):;



    #  // Function: unpack_real
    #  //
    #  // Unpacks the next 64 bits of the pack array and places them into a
    #  // real variable.
    #  //
    #  // The 64 bits of packed data are converted to a real using the $bits2real
    #  // system function.
    #
    #  extern def unpack_real(self):



    #  // Function: unpack_object
    #  //
    #  // Unpacks an object and stores the result into ~value~.
    #  //
    #  // ~value~ must be an allocated object that has enough space for the data
    #  // being unpacked. The first four bits of packed data are used to determine
    #  // if a ~null~ object was packed into the array.
    #  //
    #  // The <is_null> function can be used to peek at the next four bits in
    #  // the pack array before calling this method.
    #
    #  extern def unpack_object(self,uvm_object value):
    def unpack_object(self, value):
        is_non_null = 1

        if value in value._m_uvm_status_container.cycle_check:
            uvm_warning("CYCFND", sv.sformatf(
                "Cycle detected for object @%0d during unpack", value.get_inst_id()))
            return
        value._m_uvm_status_container.cycle_check[value] = 1

        if self.use_metadata == 1:
            is_non_null = get_bits(self.m_bits, self.count, 4) != 0  # [count +: 4]
            self.count += 4

        # NOTE- policy is a ~pack~ policy, not unpack policy;
        #       and you can't pack an object by REFERENCE
        if value is not None:
            if is_non_null > 0:
                self.scope.down(value.get_name())
                value._m_uvm_field_automation(None, UVM_UNPACK,"")
                value.do_unpack(self)
                self.scope.up()
            else:
                pass
                # TODO: help do_unpack know whether unpacked result would be null
                #       to avoid new'ing unnecessarily;
                #       this does not nullify argument; need to pass obj by ref
        elif ((is_non_null != 0) and (value is None)):
             uvm_error("UNPOBJ","cannot unpack into None object")
        del value._m_uvm_status_container.cycle_check[value]


    #  // Function: get_packed_size
    #  //
    #  // Returns the number of bits that were packed.
    def get_packed_size(self) -> int:
        return self.m_packed_size


    #  extern def unpack_object_ext(self,inout uvm_object value):


    #  extern def get_packed_bits(self):
    def get_packed_bits(self) -> int:
        return self.m_bits


    #  extern def bit  unsigned get_bit  (self,int unsigned index):
    def get_bit(self, index) -> int:
        if index >= self.m_packed_size:
            self.index_error(index, "bit",1)
        return (self.m_bits >> index) & 0x1


    #  extern def byte unsigned get_byte (self,int unsigned index):
    #  extern def int  unsigned get_int  (self,int unsigned index):

    #  extern def get_bits(self,ref bit unsigned bits[]):
    def get_bits(self) -> int:
        return self.m_bits

    #  extern def get_bytes(self,ref byte unsigned bytes[]):
    def get_bytes(self) -> List[int]:
        sz = 0
        v = 0x00
        sz = int((self.m_packed_size+7) / 8)
        bytes = [0] * sz
        for i in range(sz):
            if (i != sz-1 or (self.m_packed_size % 8) == 0):
                v = (self.m_bits >> (i * 8)) & 0xFF
            else:
                sel = (0xFF >> (8-(self.m_packed_size % 8)))
                v = (self.m_bits >> (i * 8)) & sel
            if self.big_endian:
                v = self.flip_bit_order(v, 8)
            bytes[i] = v
        return bytes


    #  extern def get_ints(self):
    def get_ints(self) -> List[int]:
        sz = 0
        v = 0
        sz = int((self.m_packed_size+31) / SIZEOF_INT)
        ints = [0] * sz
        for i in range(sz):
            if i != sz-1 or (self.m_packed_size % 32) == 0:
                v = (self.m_bits >> (i * SIZEOF_INT)) & 0xFFFFFFFF
            else:
                sel = (0xFFFFFFFF >> (32-(self.m_packed_size % 32)))
                v = (self.m_bits >> (i * SIZEOF_INT)) & sel
            if self.big_endian:
                v = self.flip_bit_order(v, SIZEOF_INT)
            ints[i] = v
        return ints


    #  extern def put_bits(self,ref bit unsigned bitstream[]):
    def put_bits(self, bitstream):
        #  int bit_size
        #  bit_size = bitstream.size()
        #
        #  if(self.big_endian)
        #    for (int i=bit_size-1;i>=0;i--)
        #      self.m_bits[i] = bitstream[i]
        #  else
        #    for (int i=0;i<bit_size;i++)
        #      self.m_bits[i] = bitstream[i]
        #
        self.m_bits = bitstream
        self.m_packed_size = len(bin(self.m_bits)) - 2
        self.count = 0

    #  extern def put_bytes(self,ref byte unsigned bytestream[]):
    #  extern def put_ints(self,ref int unsigned intstream[]):


    def set_packed_size(self):
        self.m_packed_size = self.count
        self.count = 0

    #  extern def index_error(self,int index, string id, int sz):
    def index_error(self, index, id, sz):
        uvm_error("PCKIDX",
            sv.sformatf("index %0d for get_%0s too large; valid index range is 0-%0d.",
                index,id,((self.m_packed_size+sz-1)/sz)-1))

    #  extern def enough_bits(self,int needed, string id):
    def enough_bits(self, needed, id, is_error=True):
        if ((self.m_packed_size - self.count) < needed):
            if is_error is True:
                uvm_error("PCKSZ",
                    sv.sformatf("%0d bits needed to unpack %0s, yet only %0d available.",
                    needed, id, (self.m_packed_size - self.count)))
            return 0
        return 1


    def reset(self):
        self.count = 0
        self.m_bits = 0
        self.m_packed_size = 0

    def flip_bit_order(self, value, size: int) -> int:
        flipped = 0x0
        num_bits = len(bin(value)) - 2
        while value:
            flipped = (flipped << 1) + (value & 0x1)  # Choose LSB
            value = value >> 1
        # For packing, need to add some right-padding
        if size != -1:
            rem_bits = size - num_bits
            if rem_bits >= 0:
                flipped <<= rem_bits
            else:
                raise Exception("rem_bits negative. size: {}, value: {}".format(
                    size, hex(value)))
        return flipped


#//------------------------------------------------------------------------------
#// IMPLEMENTATION
#//------------------------------------------------------------------------------
#
#// NOTE- max size limited to BITSTREAM bits parameter (default: 4096)
#
#
#// put_bytes
#// ---------
#
#def void UVMPacker::put_bytes (self,ref byte unsigned bytestream []):
#
#  int byte_size
#  int index
#  byte unsigned b
#
#  byte_size = bytestream.size()
#  index = 0
#  for (int i=0;i<byte_size;i++):
#    b = bytestream[i]
#    if(self.big_endian):
#      byte unsigned tb; tb = b
#      for(int j=0;j<8;++j) b[j] = tb[7-j]
#    end
#    self.m_bits[index +:8] = b
#    index += 8
#  end
#
#  m_packed_size = byte_size*8
#  count = 0
#endfunction
#
#
#// put_ints
#// --------
#
#def void UVMPacker::put_ints (self,ref int unsigned intstream []):
#
#  int int_size
#  int index
#  int unsigned v
#
#  int_size = intstream.size()
#
#  index = 0
#  for (int i=0;i<int_size;i++):
#    v = intstream[i]
#    if(self.big_endian):
#      v = self.flip_bit_order(v)
#    self.m_bits[index +:32] = v
#    index += 32
#  end
#
#  m_packed_size = int_size*32
#  count = 0
#endfunction
#
#
#
#
#
#
#// get_byte
#// --------
#
#def byte unsigned UVMPacker::get_byte(self,int unsigned index):
#  if (index >= (m_packed_size+7)/8)
#    index_error(index, "byte",8)
#  return self.m_bits[index*8 +: 8]
#endfunction
#
#
#// get_int
#// -------
#
#def int unsigned UVMPacker::get_int(self,int unsigned index):
#  if (index >= (m_packed_size+31)/32)
#    index_error(index, "int",32)
#  return self.m_bits[(index*32) +: 32]
#endfunction
#
#
#// PACK
#
#
#
#
#// pack_real
#// ---------
#
#def void UVMPacker::pack_real(self,real value):
#  pack_field_int($realtobits(value), 64)
#endfunction
#
#
#// pack_time
#// ---------
#
#def void UVMPacker::pack_time(self,time value):
#  pack_field_int(value, 64)
#  //m_bits[count +: 64] = value; this overwrites endian adjustments
#endfunction
#
#
#
#
#
#// pack_bits
#// -----------------
#
#def void UVMPacker::pack_bits(self,ref bit value[], input int size = -1):
#   if (size < 0)
#     size = len(value)
#
#   if (size > len(value)):
#      uvm_error("UVM/BASE/PACKER/BAD_SIZE",
#                 sv.sformatf("pack_bits called with size '%0d', which exceeds len(value) of '%0d'",
#                           size,
#                           len(value)))
#      return
#   end
#
#   for (int i=0; i<size; i++)
#     if (self.big_endian == 1)
#       self.m_bits[count+i] = value[size-1-i]
#     else
#       self.m_bits[count+i] = value[i]
#   count += size
#endfunction
#
#
#
#
#
#
#// UNPACK
#
#
#// is_null
#// -------
#
#def bit UVMPacker::is_null(self):
#  return (m_bits[count+:4]==0)
#endfunction
#
#// unpack_object
#// -------------
#
#def void UVMPacker::unpack_object_ext(self,inout uvm_object value):
#  unpack_object(value)
#endfunction
#
#
#
#// unpack_real
#// -----------
#
#def real UVMPacker::unpack_real(self):
#  if (self.enough_bits(64,"real")):
#    return $bitstoreal(unpack_field_int(64))
#  end
#endfunction
#
#
#// unpack_time
#// -----------
#
#def time UVMPacker::unpack_time(self):
#  if (self.enough_bits(64,"time")):
#    return unpack_field_int(64)
#  end
#endfunction
#
#
#
#
#
#// unpack_bits
#// -------------------
#
#def void UVMPacker::unpack_bits(self,ref bit value[], input int size = -1):
#   if (size < 0)
#     size = len(value)
#
#   if (size > len(value)):
#      uvm_error("UVM/BASE/PACKER/BAD_SIZE",
#                 sv.sformatf("unpack_bits called with size '%0d', which exceeds len(value) of '%0d'",
#                           size,
#                           len(value)))
#      return
#   end
#
#   if (self.enough_bits(size, "integral")):
#      count += size
#      for (int i=0; i<size; i++)
#        if (self.big_endian == 1)
#          value[i] = self.m_bits[count-i-1]
#        else
#          value[i] = self.m_bits[count-size+i]
#   end
#endfunction
#
#

def get_bits(bits, count, nbits):  # [count +: 4]
    val = 0x0
    idx = 0
    for i in range(count, count + nbits + 1):
        val |= (1 << (nbits - idx)) & (bits >> (i-idx))
        idx += 1
    return val




import unittest
from uvm.reg.uvm_reg_model import *


class TestUVMRegModel(unittest.TestCase):


    def test_concat2string(self):
        concat = uvm_hdl_path_concat()
        path_sli = uvm_hdl_path_slice()
        path_sli.path = "dut"
        concat.add_path(path_sli.path)
        str_path = uvm_hdl_concat2string(concat)
        self.assertEqual(str_path, "dut")
        concat.add_path("test_reg")
        str_path = uvm_hdl_concat2string(concat)
        self.assertEqual(str_path, "{dut, test_reg}")


if __name__ == '__main__':
    unittest.main()
